# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports
import time
import os
import subprocess
import sys

# Other imports
import xml.etree.ElementTree

import numpy as np
import re
from datetime import datetime
from pathlib import Path
from collections import OrderedDict

import scipy.ndimage
from silx.io.specfile import SpecFile

import scipy.ndimage as ndimage
from scipy.signal import medfilt2d

import pandas as pd
import yaml
import json
import h5py
import fabio

# This module imports
from .lmfit_models import PlaneModel, Gaussian2DModel, LorentzianSquared2DModel, Pvoigt2DModel, update_param_hints

from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

# Detector File Sizes
detector_file_sizes = {
    'Rayonix MX225': 18878464,
    'Rayonix SX165': 8392704,
    'Pilatus 100k': 379860,
    'Pilatus 1M': 4092732,
}


def write_xye(fname, xdata, ydata, variance=None):
    """Saves data to an xye file. Variance is the square root of the
    signal.
    
    args:
        fname: str, path to file
        xdata: angle or q data
        ydata: intensity
    """
    if variance is None:
        _variance = np.sqrt(abs(ydata))
    else:
        _variance = variance
    with open(fname, "w") as file:
        for i in range(0, len(xdata)):
            file.write(
                str(xdata[i]) + "\t" +
                str(ydata[i]) + "\t" +
                str(_variance[i]) + "\n"
            )


def write_csv(fname, xdata, ydata, variance=None):
    """Saves data to a csv file.
    
    args:
        fname: str, path to file
        xdata: angle or q data
        ydata: intensity
    """
    if variance is None:
        _variance = np.sqrt(abs(ydata))
    else:
        _variance = variance
    with open(fname, 'w') as file:
        for i in range(0, len(xdata)):
            file.write(str(xdata[i]) + ', ' +
                       str(ydata[i]) + ', ' +
                       str(_variance[i]) + '\n')


def check_encoded(grp, name):
    """Checks grp attributes for encoded, checks resulting attribute
    against name. If encoded not an attribute, returns False.
    """
    return grp.attrs.get("encoded", "not_found") == name


def find_between( s, first, last ):
    """find first occurence of substring in string s
     between two substrings (first and last)

    Args:
        s (str): input string
        first (str): first substring
        last (str): second substring

    Returns:
        str: substring between first and last
    """
    try:
        start = s.index( first ) + len( first )
        end = s.index( last, start )
        return s[start:end]
    except ValueError:
        return ""


def find_between_r( s, first, last ):
    """find last occurence of substring in string s
     between two substrings (first and last)

    Args:
        s (str): input string
        first (str): first substring
        last (str): second substring

    Returns:
        str: substring between first and last
    """
    try:
        start = s.rindex(first) + len(first)
        end = s.rindex(last, start)
        return s[start:end]
    except ValueError:
        return ""


# def get_fname_dir(fname):
def get_fname_dir():
    """
    Returns directory on local drive to save temporary h5 files in

    Args:
        fname: {str} Name of scan name used to create subdirectory

    Returns:
        path: {str} Path where h5 file is saved
    """
    home_path = str(Path.home())

    path = os.path.join(home_path, 'xdart_processed_data')
    Path(path).mkdir(parents=True, exist_ok=True)

    return path


def split_file_name(fname):
    """Splits filename to get directory, file root and extension

    Arguments:
        fname {str} -- full image file name with path
    """
    directory = os.path.dirname(fname)
    root, ext = os.path.splitext(os.path.basename(fname))

    if len(ext) > 0:
        if ext[0] == '.':
            ext = ext[1:]

    return directory, root, ext


def get_sname_img_number(fname):
    """Splits filename to get scan name and image number

    Arguments:
        fname {str} -- full image file name with path
    Returns:
        series_name {str} -- series name (or root if single acquisition)
        img_number {int/None} -- image number (if part of series)
    """
    directory, root, ext = split_file_name(fname)
    try:
        img_number = int(root[root.rindex('_') + 1:])
        root = root[:root.rindex('_')]
    except ValueError:
        img_number = None

    return root, img_number


def get_series_avg(fname, detector, meta_ext):
    """ Returns the averaged image and meta data for a series

    Arguments:
        fname {str} -- full image file name with path
        detector {obj} -- pyFAI detector object
    Returns:
        series_name {str} -- series name (if series exists)
        series_file_names {str array} -- file names in series
        img_data {ndarray} -- averaged 2D intensity over a series
        img_mete {dict} -- averaged meta data over a series
    """
    fpath = Path(fname)
    series_name, img_number = get_sname_img_number(fname)
    if img_number is None:
        return None, None, None, None

    fnames = [str(f) for f in (fpath.parent.glob(f'{series_name}*[0-9][0-9][0-9][0-9]{fpath.suffix}'))]
    data, img_meta = 0, {}
    for ii, fname in enumerate(fnames):
        data += get_img_data(fname, detector, return_float=True)
        if data is None:
            return None, None, None, None

        meta = get_img_meta(fname, meta_ext) if meta_ext else {}
        if ii == 0:
            img_meta = meta
        else:
            for (k, v) in meta.items():
                try:
                    img_meta[k] += meta[k]
                except TypeError:
                    pass

    n = len(fnames)
    data /= n
    for (k, v) in img_meta.items():
        try:
            img_meta[k] /= n
        except TypeError:
            pass

    return series_name, fnames, data, img_meta


def get_scan_name(fname):
    """Splits filename to get scan name

    Arguments:
        fname {str} -- full image file name with path
    Returns:
        scan_name {str}
    """
    directory, root, ext = split_file_name(fname)
    try:
        img_number = root[root.rindex('_') + 1:]
    except ValueError:
        img_number = ''

    if img_number:
        try:
            _ = int(img_number)
            return root[:root.rindex('_')]
        except ValueError:
            return root

    return root


def get_img_number(fname):
    """Splits filename to get scan name and image number

    Arguments:
        fname {str} -- full image file name with path
    Returns:
        scan_name {str}
        nImage {int}
    """
    # directory, root, ext = split_file_name(fname)
    root = os.path.splitext(fname)[0]
    try:
        img_number = root[root.rindex('_') + 1:]
        img_number = int(img_number)
    except ValueError:
        img_number = 1

    return img_number


def match_img_detector(img_file, poni_dict):
    """Check if the file is created by the detector specified"""
    detector_name = poni_dict['detector'].name
    if detector_name not in detector_file_sizes.keys():
        return True

    if os.stat(img_file).st_size == detector_file_sizes[detector_name]:
        return True

    return False


def query(question):
    """Ask a question with allowed options via input()
    and return their answer.
    """
    sys.stdout.write(question)
    return input()


def get_spec_file(img_fname):
    """Check if SPEC file exists for an image file and return path if yes"""
    fpath = Path(img_fname)
    fname = fpath.stem
    match = re.search(f'_scan\d+_\d+.', fname)
    if match is None:
        return None
    spec_fname = fname[fname.find('_') + 1:match.start()]

    for nn in range(3):
        s = os.path.join(fpath.parents[nn], spec_fname)
        if os.path.isfile(s):
            return s

    return None


def get_img_meta(img_file, meta_ext, spec_path=None, rv='all'):
    """Get image meta data from pdi/txt files for different beamlines

    Args:
        img_file (str): Image file for which meta data is required
        meta_ext (str): Extension of Meta file
        spec_path (str): Path of spec file if not in regular path
        rv (str, optional): Return values (Counters, motors or all)

    Returns:
        [dict]: Dictionary with all the meta data
    """
    if meta_ext == 'SPEC':
        Counters, Motors, Extras = get_meta_from_spec(img_file, spec_path)
    else:
        meta_file = f'{os.path.splitext(img_file)[0]}.{meta_ext}'
        meta_file = meta_file if os.path.exists(meta_file) else f'{img_file}.{meta_ext}'

        if not os.path.exists(meta_file):
            return {}

        if meta_ext == 'pdi':  # Pilatus Image
            Counters, Motors, Extras = get_meta_from_pdi(meta_file)
        else:
            Counters, Motors, Extras = get_meta_from_txt(meta_file)

    if rv == 'Counters':
        return Counters
    elif rv == 'Motors':
        return Motors

    return Extras | Counters | Motors


def get_meta_from_spec(img_file, spec_path=None, spec_file=None, img_number=None):
    if spec_file is None:
        spec_file = get_specFile_scanNumber(img_file, spec_path)
    if spec_file is None:
        return {}, {}, {}

    scan_number = get_scanNumber(img_file)

    if img_number is None:
        img_number = get_img_number(img_file)

    ic(spec_file)
    sf_object = SpecFile(spec_file)
    sf = sf_object[scan_number - 1]
    try:
        Counters = {c: v for c, v in zip(sf.labels, sf.data_line(img_number))}
        Motors = {m: v for m, v in zip(sf.motor_names, sf.motor_positions)}
    except IndexError:
        Counters, Motors = {}, {}
    sf_object.close()

    Extras = {}
    return Counters, Motors, Extras


def get_scanNumber(img_file):
    img_fname = Path(img_file).stem

    match = re.search(f'_scan\d+_\d+$', img_fname)
    if match is None:
        return None

    return int(img_fname[match.start() + 5: img_fname.rfind('_')])


def get_specFile(img_file, spec_path=None):
    img_fname = Path(img_file).stem
    if img_fname[0:2] == 'b_':
        img_fname = img_fname[2:]

    match = re.search(f'_scan\d+_\d+$', img_fname)
    if match is None:
        return None

    spec_fname = img_fname[img_fname.find('_') + 1:match.start()]

    if spec_path is not None:
        s = os.path.join(spec_path, spec_fname)
        if os.path.exists(s):
            return s
    else:
        img_fpath = Path(img_file)
        for nn in range(2):
            s = os.path.join(img_fpath.parents[nn], spec_fname)
            ic(s)
            if os.path.isfile(s):
                return s

    return None


def get_specFile_scanNumber(img_file, spec_path=None):
    img_file = Path(img_file)
    img_fname = os.path.basename(img_file)
    if img_fname[0:2] == 'b_':
        img_fname = img_fname[2:]
    img_ext = img_file.suffix[1:]

    match = re.search(f'_scan\d+_\d+.{img_ext}', img_fname)
    if match is None:
        return None, None

    spec_fname = img_fname[img_fname.find('_') + 1:match.start()]
    scan_number = int(img_fname[match.start() + 5: img_fname.rfind('_')])
    # print(f'{img_fname} \n{spec_fname} \n{match.group()} \n{img_ext} \n{scan_number}')

    if spec_path is not None:
        s = os.path.join(spec_path, spec_fname)
        if os.path.exists(s):
            return s, scan_number
    else:
        for nn in range(2):
            s = os.path.join(img_file.parents[nn], spec_fname)
            if os.path.isfile(s):
                return s, scan_number

    return None, None


def get_meta_from_pdi(pdi_file):
    """Get motor and counter names and values from PDI file

    Args:
        pdi_file (str): PDI file name with path

    Returns:
        [dict]: Tuple of two dictionaries containing Counters and Motors
    """
    with open(pdi_file, 'r') as f:
        data = f.read()
    data = data.replace('\n', ';')

    try:
        counters = re.search('All Counters;(.*);;# All Motors', data).group(1)
        cts = re.split(';|=', counters)
        Counters = {c.split()[0]: float(cs) for c, cs in zip(cts[::2], cts[1::2])}

        motors = re.search('All Motors;(.*);#', data).group(1)
        cts = re.split(';|=', motors)
        Motors = {c.split()[0]: float(cs) for c, cs in zip(cts[::2], cts[1::2])}
    except AttributeError:
        ss1 = '# Diffractometer Motor Positions for image;# '
        ss2 = ';# Calculated Detector Calibration Parameters for image:'

        try:
            motors = re.search(f'{ss1}(.*){ss2}', data).group(1)
            cts = re.split(';|=', motors)
            Motors = {c.split()[0]: float(cs) for c, cs in zip(cts[::2], cts[1::2])}
            Motors['TwoTheta'] = Motors['2Theta']
        except AttributeError:
            Motors = {'TwoTheta': float(0.0), 'Theta': float(0.0)}
        Counters = {}

    Extras = {}
    if len(data[data.rindex(';') + 1:]) > 0:
        Extras['epoch'] = data[data.rindex(';') + 1:]

    return Counters, Motors, Extras


def get_meta_from_txt(txt_file):
    """Get motor and counter names and values from PDI file

    Args:
        txt_file (str): Txt meta file name with path

    Returns:
        [dict]: Tuple of two dictionaries containing Counters and Motors
    """
    with open(txt_file, 'r') as f:
        data = f.read()

    counters = re.search('# Counters\n(.*)\n', data).group(1)
    cts = re.split(',|=', counters)
    Counters = {c.split()[0]: float(cs) for c, cs in zip(cts[::2], cts[1::2])}

    motors = re.search('# Motors\n(.*)\n', data).group(1)
    cts = re.split(',|=', motors)
    Motors = {c.split()[0]: float(cs) for c, cs in zip(cts[::2], cts[1::2])}

    # image_meta_data['User'] = (data[data.index('User: ')+5: data.index(', time')]).strip()
    # image_meta_data['Time'] = (data[data.index('time: ')+5: data.index('# Temp')-2]).strip()
    Time = (data[data.index('time: ') + 5: data.index('# Temp') - 2]).strip()

    # d = datetime.strptime(image_meta_data['Time'], "%a %b %d %H:%M:%S %Y")
    Extras = {}
    d = datetime.strptime(Time, "%a %b %d %H:%M:%S %Y")
    Extras['epoch'] = time.mktime(d.timetuple())

    return Counters, Motors, Extras


def get_motor_val(pdi_file, motor):
    """Return position of a particular motor from PDI file

    Args:
        pdi_file (str): PDI file name with path
        motor (str): Motor name

    Returns:
        float: Motor position
    """
    _, Motors = get_meta_from_pdi(pdi_file)

    return Motors[motor]


def get_img_data(
        fname, detector, orientation='horizontal',
        flip=False, fliplr=False, transpose=False,
        return_float=False, im=0):
    """Read image file and return numpy array

    Args:
        fname (str): File Name with path
        detector (detector object): pyFAI detector object
        orientation (str, optional): Orientation of detector. Options: 'horizontal', 'vertical'. Defaults to 'horizontal'.
        flip (bool, optional): Flag to flip the image up-down (required by pyFAI at times). Defaults to False.
        fliplr (bool, optional): Flag to flip the image left-right (required by pyFAI at times). Defaults to False.
        transpose (bool, optional): Flag to transpose the image (required by pyFAI at times). Defaults to False.
        return_float (bool, optional): Convert array to float. Defaults to False.
        im (integer, optional): image number if input is h5 file from Eiger. Defaults to 0

    Returns:
        ndarray: Image data read into numpy array
    """
    try:
        img_data = fabio.open(fname).data
    except xml.etree.ElementTree.ParseError:
        return None
    except OSError:
        if detector.name == 'Eiger 1M':
            with h5py.File(fname, mode='r') as f:
                try:
                    img_data = np.asarray(f['entry']['data']['data'][im], dtype=float)
                except IndexError:
                    return None
        else:
            img_data = np.asarray(np.fromfile(fname, dtype='int32', sep=""), dtype=float)
            try:
                img_data = img_data.reshape(detector.shape)
            except ValueError:
                return None
    except ValueError:
        return None
    except:
        return None

    try:
        if img_data.shape != detector.shape:
            return None
    except AttributeError:
        return None

        # if 'tif' in fname[-5:]:
        #     img_data = np.asarray(io.imread(fname))
        # elif ('h5' in fname[-4:]) or ('hdf5' in fname[-6:]):
        #     with h5py.File(fname, mode='r') as f:
        #         try:
        #             img_data = np.asarray(f['entry']['data']['data'][im], dtype=float)
        #         except IndexError:
        #             return None
        #         img_data[514:551, :] = np.nan
        #
        #         # Hot pixel in SSRL Eiger 1M detector
        #         img_data[0, 1029] = np.nan
        # elif 'mar3450' in fname[-9:]:
        #     img_data = fabio.open(fname).data
        # else:
        #     img_data = np.asarray(np.fromfile(fname, dtype='int32', sep=""), dtype=float)
        #     if len(img_data) == np.prod(shape_100K):
        #         img_data = img_data.reshape(shape_100K)
        #     elif len(img_data) == np.prod(shape_300K):
        #         img_data = img_data.reshape(shape_300K)
        #     else:
        #         img_data = img_data.reshape(shape_1M)
        #         img_data[:, 487:487 + 7] = np.nan
        #         for ii in range(1, 5):
        #             mod_start = 195 * ii + 17 * (ii - 1)
        #             img_data[mod_start:mod_start + 17] = np.nan

    if return_float:
        img_data = np.asarray(img_data, dtype=float)

    if (orientation == 'vertical') or transpose:
        img_data = img_data.T

    if flip:
        img_data = np.flipud(img_data)

    if fliplr:
        img_data = np.fliplr(img_data)

    return img_data


def get_mask_array(detector, mask_file=None, det_orientation=0):
    """Get mask array from mask file
    """
    mask = detector.calc_mask()
    if mask_file and os.path.exists(mask_file):
        if mask is not None:
            try:
                mask += fabio.open(mask_file).data
            except ValueError:
                print('Mask file not valid for Detector')
                pass
        else:
            mask = fabio.open(mask_file).data

    if mask is None:
        return None

    if mask.shape != detector.shape:
        print('Mask file not valid for Detector')
        return None

    mask = scipy.ndimage.rotate(mask, det_orientation)
    return np.flatnonzero(mask)


def get_norm_fac(normChannel, scan_data, arch_ids=None, return_sum=True):
    """Check to see if normalization channel exists in metadata and return name"""
    normChannel = get_normChannel(normChannel, scan_data_keys=scan_data.columns)
    if arch_ids is None:
        arch_ids = scan_data.index
    norm_fac = scan_data[normChannel][arch_ids] if normChannel else 1
    if return_sum and not isinstance(norm_fac, int):
        norm_fac = norm_fac.mean()

    return norm_fac


def get_normChannel(normChannel, scan_data_keys):
    """Check to see if normalization channel exists in metadata and return name"""
    if normChannel == 'sec':
        normChannel = {'sec', 'seconds', 'Seconds', 'Sec', 'SECONDS', 'SEC'}
    else:
        normChannel = {normChannel, normChannel.lower(), normChannel.upper()}
    normChannel = normChannel.intersection(scan_data_keys)
    return normChannel.pop() if len(normChannel) > 0 else None


def smooth_img(img, kernel_size=3, window_size=3, order=0):
    """Apply a Gaussian filter to smooth image

    Args:
        img (ndarray): 2D numpy array for image
        kernel_size (int, optional): Gaussian filter kernel. Defaults to 3.
        window_size (int, optional): Gaussian filter window (should be odd). Defaults to 3.
        order (int, optional): Order of the filter. Defaults to 0.

    Returns:
        ndarray: Smoothed image
    """
    if (np.mod(kernel_size, 2) == 0) or (np.mod(window_size, 2) == 0):
        print('Smoothing windows should be odd integers')
        return img

    if order >= window_size:
        order = window_size - 1

    if kernel_size > 1:
        img = medfilt2d(img, 3)
    if window_size > 1:
        img = ndimage.gaussian_filter(img, sigma=(window_size, window_size), order=order)

    return img


def get_fit(im, function='gaussian'):
    """Custom function to perform 2D fit (using lmfit) on an image

    Args:
        im (ndarray): 2D array representing the image
        function (str, optional): Fitting function to use. Defaults to 'gaussian'.

    Returns:
        tuple: Fit result (lmfit Model object), fit parameters, and evaluated fit
    """
    # Flatten Arrays
    ydata = im.flatten()
    nrows, ncols = im.shape
    rows, cols = np.meshgrid(np.arange(0,ncols), np.arange(0,nrows))
    x = [rows.flatten(), cols.flatten()]

    Models = {'gaussian': Gaussian2DModel(),
              'lorentzian': LorentzianSquared2DModel(),
              'pvoigt': Pvoigt2DModel()}
    
    plane_mod = PlaneModel()
    curve_mod = Models[function]
    #curve_mod = Gaussian2DModel()

    # Using the guess function
    pars = plane_mod.guess(ydata, x)
    pars += curve_mod.guess(ydata, x)

    Hints = {'sigma_x':   {'value':5, 'min': 1},
             'sigma_y':   {'value':5, 'min': 1},
             'amplitude': {'min': 0},
             'intercept': {'value':0, 'vary':False},
             'slope_x':   {'value':0, 'vary':False},
             'slope_y':   {'value':0, 'vary':False},
            }

    update_param_hints(pars, **Hints)
    
    mod_2D = plane_mod + curve_mod
    mod_2D.missing = 'drop'

    out = mod_2D.fit(ydata, pars, x=x)

    # Fit results
    I_fit = out.eval(x=[rows.flatten(), cols.flatten()])
    im_fit = I_fit.reshape(im.shape)
    
    return out, pars, im_fit


def fit_images_2D(fname, detector, tth, function='gaussian',
                  kernel_size=3, window_size=3, order=0,
                  Fit_Results={}, FNames={}, Img_Fits={}, Init_Params={},
                  verbose=False, **kwargs):
    """Wrapper function aroung get_fit function

    Args:
        fname (str): Image file name
        detector (detector object): pyFAI detector object
        tth (float): Value of 2th (used as key for returned dictionary)
        function (str, optional): Fitting function. Defaults to 'gaussian'.
        kernel_size (int, optional): Gaussian smoothing kernel size. Defaults to 3.
        window_size (int, optional): Gaussian smoothing window size. Defaults to 3.
        order (int, optional): Order of Gaussian filter. Defaults to 0.
        Fit_Results (dict, optional): Dictionary with tth as key containing fit result. Defaults to {}.
        FNames (dict, optional): Dictionary containing file names for each tth. Defaults to {}.
        Img_Fits (dict, optional): Dictionary containing the evaluated fits for each tth. Defaults to {}.
        Init_Params (dict, optional): Dictionary containing initial fit parameters for each tth. Defaults to {}.
        verbose (bool, optional): Flag to print debug messages. Defaults to False.

    Returns:
        tuple: tth value and fit result
    """
    if verbose:
        print(f'Processing {fname}')
    img = get_img_data(fname, detector, return_float=True, verbose=False, **kwargs)
    
    smooth_img(img, kernel_size=kernel_size, window_size=window_size, order=order)
    fit_result, init_params, img_fit = get_fit(img, function=function)
    
    tth = np.round(tth, 3)
    Fit_Results[tth] = fit_result
    FNames[tth] = fname
    Img_Fits[tth] = img_fit
    Init_Params[tth] = init_params
    
    return (tth, fit_result)


def data_to_h5(data, grp, key, encoder='yaml', compression='lzf'):
    """Saves data to hdf5 file. Global function, calls other functions
    based on data type. If data type can't be determined, tries to
    save as a dataset, and ultimately defaults to a string
    representation.
    
    args:
        data: object to be saved
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        key: str, name of new Group or Dataset
        encoder: str, 'yaml' or 'json', how to encode stubborn data
            types.
        compression: str, compression algorithm to use. See h5py docs.
    """
    if data is None:
        none_to_h5(grp, key)

    elif type(data) == dict:
        dict_to_h5(data, grp, key, compression=compression)
    
    elif type(data) == str:
        str_to_h5(data, grp, key)
    
    elif type(data) == pd.core.series.Series:
        series_to_h5(data, grp, key, compression)
    
    elif type(data) == pd.core.frame.DataFrame:
        dataframe_to_h5(data, grp, key, compression)

    else:
        try:
            if np.array(data).shape == ():
                scalar_to_h5(data, grp, key)
            else:
                arr_to_h5(data, grp, key, compression)

        except TypeError:
            try:
                encoded_h5(data, grp, key, encoder)
            except Exception as e:
                print(e)
                try:
                    if key in grp:
                        if check_encoded(grp[key], 'unknown'):
                            grp[key][()] = np.string_(data)
                            return
                        else:
                            del(grp[key])
                    grp.create_dataset(key, data=np.string_(data),
                                    dtype=h5py.string_dtype())
                    grp[key].attrs['encoded'] = 'unknown'
                except Exception as e:
                    print(e)
                    print(f"Unable to dump {key}")


def none_to_h5(grp, key):
    if key in grp:
        del(grp[key])
    grp.create_dataset(key, data=h5py.Empty("f"))
    grp[key].attrs['encoded'] = 'None'


def dict_to_h5(data, grp, key, **kwargs):
    """Adds dictionary data to hdf5 group with same keys as dictionary.
    See data_to_h5 for how datatypes are handled.

    args:
        data: dict, dictionary to add to hdf5
        grp: h5py Group or File, where to add the data to
        key: str, name of new group
        **kwargs: passed on to data_to_h5.
    """
    if key in grp:
        if not check_encoded(grp[key], "dict"):
            del(grp[key])
            new_grp = grp.create_group(key)
            new_grp.attrs["encoded"] = "dict"
        else:
            new_grp = grp[key]
    else:
        new_grp = grp.create_group(key)
    
    for jey in data:
        s_key = str(jey)
        sub_data = data[jey]
        data_to_h5(sub_data, new_grp, s_key, **kwargs)


def str_to_h5(data, grp, key):
    """Saves string to hdf5 file. Saved as h5py.string_dtype, if
    key exists will overwrite.
    
    args:
        data: str, object to be saved
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        key: str, name of new Group or Dataset
    """
    if key in grp:
        if check_encoded(grp[key], "str"):
            grp[key][()] = data
        else:
            del(grp[key])
            grp.create_dataset(key, data=data, dtype=h5py.string_dtype())
    else:
            grp.create_dataset(key, data=data, dtype=h5py.string_dtype())


def series_to_h5(data, grp, key, compression):
    """Saves pandas Series to hdf5 file.
    
    args:
        data: Series, object to be saved
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        key: str, name of new Group or Dataset
        compression: str, compression algorithm to use. See h5py docs.
    """
    if key in grp:
        if check_encoded(grp[key], "Series"):
            new_grp = grp[key]
            new_grp['data'][()] = np.array(data)
            index_to_h5(data.index, 'index', new_grp, compression)
            new_grp.attrs['name'] = data.name
            return
        else:
            del(grp[key])
    
    new_grp = grp.create_group(key)
    new_grp.attrs['encoded'] = 'Series'
    new_grp.create_dataset('data', data=np.array(data), compression=compression,
                           chunks=True)
    index_to_h5(data.index, 'index', new_grp, compression)
    new_grp.attrs['name'] = data.name


def dataframe_to_h5(data, grp, key, compression):
    """Saves pandas DataFrame to hdf5 file.
    
    args:
        data: DataFrame, object to be saved
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        key: str, name of new Group or Dataset
        compression: str, compression algorithm to use. See h5py docs.
    """
    if key in grp:
        if check_encoded(grp[key], "DataFrame"):
            new_grp = grp[key]
        else:
            del(grp[key])
            new_grp = grp.create_group(key)
            new_grp.attrs['encoded'] = 'DataFrame'
    else:
        new_grp = grp.create_group(key)
        new_grp.attrs['encoded'] = 'DataFrame'
    index_to_h5(data.index, 'index', new_grp, compression)
    index_to_h5(data.columns, 'columns', new_grp, compression)
    if 'data' in new_grp:
        new_grp['data'].resize(np.array(data).shape)
        new_grp['data'][()] = np.array(data)[()]
    else:
        new_grp.create_dataset('data', data=np.array(data), compression=compression,
                                chunks=True, maxshape=(None,None))


def index_to_h5(index, key, grp, compression):
    """Saves index from Series or DataFrame to hdf5 file. If not a
    scalar index, saves data as strings.
    
    args:
        index: object to be saved
        key: str, name of new Group or Dataset
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        compression: str, compression algorithm to use. See h5py docs.
    """
    if key in grp:
        if grp[key].shape == (0,):
            del(grp[key])
        
    if index.dtype == 'object':
        if len(index) > 0:
            strindex = np.array([np.string_(str(x)) for x in index])
            if key in grp:
                grp[key].resize(strindex.shape)
                grp[key][()] = strindex[()]
            else:
                grp.create_dataset(
                    key, data=strindex, dtype=h5py.string_dtype(),
                    chunks=True, maxshape=(None,)
                )
        else:
            if key in grp:
                del(grp[key])
            grp.create_dataset(key, data=np.array([]))
    else:
        arrindex = np.array(index)
        if key in grp:
            grp[key].resize(arrindex.shape)
            grp[key][()] = arrindex[()]
        else:
            grp.create_dataset(
                key, data=np.array(index), compression=compression,
                chunks=True, maxshape=(None,)
            )


def scalar_to_h5(data, grp, key):
    """Saves scalar to hdf5 file.
    
    args:
        data: scalar, object to be saved
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        key: str, name of new Group or Dataset
    """
    if key in grp:
        if check_encoded(grp[key], 'scalar'):
            if grp[key].dtype == np.array(data).dtype:
                grp[key][()] = data
                return
        del(grp[key])
    grp.create_dataset(key, data=data)
    grp[key].attrs['encoded'] = 'scalar'


def arr_to_h5(data, grp, key, compression):
    """Saves numpy array or list to hdf5 file. Ensures resizability, and
    overwrites data present without deleting data.
    
    args:
        data: numpy array or list, object to be saved
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        key: str, name of new Group or Dataset
        compression: str, compression algorithm to use. See h5py docs.
    """
    if key in ['map_raw', 'bg_raw']:
        arr = np.array(data, dtype='int32')
    elif key in ['i_tthChi', 'i_qChi', 'i_QxyQz']:
        arr = np.array(data, dtype='float32')
    else:
        arr = np.array(data)

    if key in grp:
        if check_encoded(grp[key], 'arr'):
            if grp[key].dtype == arr.dtype:
                grp[key].resize(arr.shape)
                grp[key][()] = arr[()]
                return
        del(grp[key])
    # grp.create_dataset(key, data=arr, compression=compression, chunks=True,
    #                    maxshape=tuple(None for x in arr.shape))
    grp.create_dataset(key, data=arr, maxshape=tuple(None for x in arr.shape))
    grp[key].attrs['encoded'] = 'arr'


def encoded_h5(data, grp, key, encoder):
    """Saves data of unparsed type to hdf5 file. Uses the encoder to
    represent the data as a string.
    
    args:
        data: data to be saved, unknown type.
        grp: h5py File or Group, where data will be saved.
        key: str, name of new Dataset
        encoder: str, 'yaml' or 'json', how to encode stubborn data
            types.
    """
    if encoder == 'yaml':
        string = np.string_(yaml.dump(data))
    elif encoder == 'json':
        string = np.string_(json.dumps(data))
    if key in grp:
        if check_encoded(grp[key], encoder):
            grp[key][()] = string
            return
        else:
            del(grp[key])
    grp.create_dataset(key, data=string, dtype=h5py.string_dtype())
    grp[key].attrs['encoded'] = encoder


def attributes_to_h5(obj, grp, lst_attr=None, priv=False, dpriv=False,
                     **kwargs):
    """Function which takes a list of class attributes and stores them
    in a provided h5py group. See data_to_h5 for how datatypes are
    handled.
    
    args:
        obj: object to store
        grp: h5py File or Group, where data will be saved. Creates new
            Group or Dataset in grp.
        lst_attr: list or None, if list only the listed attributes are
            stored.
        priv: bool, if True and lst_attr is None, stores functions
            beginning with single _
        drpiv: bool, if True and lst_attr is None, stores attributes
            beginning with double _
        kwargs: passed to data_to_h5
    """
    if lst_attr is None:
        if dpriv:
            lst_attr = list(obj.__dict__.keys())
        elif priv:
            lst_attr = [x for x in obj.__dict__.keys() if '__' not in x]
        else:
            lst_attr = [x for x in obj.__dict__.keys() if '_' not in x]
    for attr in lst_attr:
        data = getattr(obj, attr)
        data_to_h5(data, grp, attr, **kwargs)


def h5_to_data(grp, encoder=True, Loader=yaml.UnsafeLoader):
    """Reads data from hdf5 file and returns appropriate datatype.
    
    args:
        grp: h5py File or Group, where the data is stored
        encoder: bool, if True checks the 'encoded' attribute to know
            the type
        Loader: yaml Loader, passed to yaml if encoded with yaml.
    
    returns:
        data: unknown type, the data in grp.
    """
    if encoder and 'encoded' in grp.attrs:
        encoded = grp.attrs['encoded']
        if encoded == 'None':
            data = None

        elif encoded == 'dict':
            data = h5_to_dict(grp, encoder=encoder, Loader=Loader)

        elif encoded == 'str':
            try:
                data = grp[...].item().decode()
            except AttributeError:
                data = grp[()]
        
        elif encoded == 'Series':
            data = pd.Series(
                data = grp['data'][()],
                index = h5_to_index(grp['index']),
                name = grp.attrs['name']
            )

        elif encoded == 'DataFrame':
            data = pd.DataFrame(
                data = grp['data'][()],
                index = h5_to_index(grp['index']),
                columns = h5_to_index(grp['columns']),
            )

        elif encoded in ['data', 'arr', 'scalar']:
            data = grp[()]

        elif encoded == 'yaml':
            data = yaml.load(grp[...].item(), Loader=Loader)

        elif encoded == 'json':
            data = json.loads(grp[...].item())

        elif encoded == 'unknown':
            try:
                data = eval(grp[...].item())
            except:
                try:
                    data = grp[...].item().decode()
                except AttributeError:
                    data = grp[...].item()
    else:
        if type(grp) == h5py._hl.group.Group:
            data = h5_to_dict(grp, encoder=encoder, Loader=Loader)

        elif grp.shape == ():
            temp = grp[...].item()
            if type(temp) == bytes:
                temp = temp.decode()
            if temp == 'None':
                data = None
            else:
                data = temp

        elif grp.shape is None:
            data = None

        else:
            data = grp[()]

    return data


def h5_to_index(grp):
    """Gets index from grp for Series or DataFrame. Uses soft_list_eval
    to handle object index types.
    
    args:
        grp: h5py File or Group, where the data is stored
        
    returns:
        list or array to be used as index
    """
    if np.issubdtype(grp.dtype, np.number):
        return grp[()]
    else:
        return soft_list_eval(grp)


def h5_to_dict(grp, **kwargs):
    """Converts h5py group to dictionary. See h5_to_data for how
    different datatypes are handled.

    args:
        grp: h5py group object

    returns:
        data: dictionary of data from h5py group
    """
    data = {}
    for key in grp.keys():
        try:
            e_key = eval(key, {})
        except:
            e_key = key

        data[e_key] = h5_to_data(grp[key], **kwargs)

    return data


def h5_to_attributes(obj, grp, lst_attr=None, **kwargs):
    """Sets attributes of obj using data in an hdf5 file. See h5_to_data
    for how data types are handled.
    
    args:
        obj: object to set attributes to
        grp: h5py File or Group, where the data is stored
        lst_attr: list of attributes to set. If not provided, sets
            all keys in grp
        kwargs: Passed to h5_to_data
    """
    if lst_attr is None:
        lst_attr = grp.keys()
    for attr in lst_attr:
        if attr in obj.__dict__.keys():
            try:
                data = h5_to_data(grp[attr], **kwargs)
                setattr(obj, attr, data)
            except KeyError:
                pass


def div0( a, b ):
    """ ignore / 0, div0( [-1, 0, 1], 0 ) -> [0, 0, 0] """
    with np.errstate(divide='ignore', invalid='ignore'):
        c = np.true_divide( a, b )
        c[ ~ np.isfinite( c )] = 0  # -inf inf NaN
    return c


def soft_list_eval(data, scope={}):
    """Tries to create list of evaluated items in data. If exception
    is thrown by eval, it just adds the element as is to the list.

    args:
        data: list or array-like, input data to be evaluated
        scope: dict, scope of functions passed to eval. See eval
            documentation

    returns:
        out: list of values in data with eval applied if possible
    """
    out = []
    for x in data:
        try:
            out.append(eval(x, scope))
        except:
            try:
                out.append(x.decode())
            except (AttributeError, SyntaxError):
                out.append(x)

    return out


def catch_h5py_file(filename, mode='r', tries=100, *args, **kwargs):
    """Forces an h5py object to be opened. Catches OSErrors which can
    be thrown. Will try a set number of times before giving up.
    
    args:
        filename: str, path to file
        mode: str, mode to open file. See h5py docs
        tries: int, how many times to try opening the file
        args, kwargs: passed to h5py.File
    """
    failed = True
    for i in range(tries):
        if i % 10 == 0 and i > 0:
            print(f"Tried catching {i} times.")
        try:
            hdf5_file = h5py.File(filename, mode, *args, **kwargs)
            failed = False
            break  # Success!
        except OSError:
            time.sleep(0.05)
            pass
    if failed:
        hdf5_file = h5py.File(filename, mode, *args, **kwargs)
    return hdf5_file


def query_yes_no(question, default="no"):
    """Ask a yes/no question via raw_input() and return their answer.
    
    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is one of "yes" or "no".
    """
    valid = {"yes":"yes",   "y":"yes",  "ye":"yes",
             "no":"no",     "n":"no"}
    if default == None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while 1:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return default
        elif choice in valid.keys():
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "\
                             "(or 'y' or 'n').\n")


class FixSizeOrderedDict(OrderedDict):
    def __init__(self, *args, max=0, **kwargs):
        self._max = max
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if self._max > 0:
            if len(self) >= self._max:
                keys = list(self.keys())
                try:
                    k = int(key)
                    diffs = [abs(int(k_)-k) for k_ in keys]
                    out_key = keys[diffs.index(max(diffs))]
                    self.pop(out_key)
                    # pos = False if (abs(k - keys[0]) > abs(k - keys[-1])) else True
                except ValueError:
                    self.popitem(False)

        OrderedDict.__setitem__(self, key, value)


def launch(program):
    """launch(program)
      Run program as if it had been double-clicked in Finder, Explorer,
      Nautilus, etc. On OS X, the program should be a .app bundle, not a
      UNIX executable. When used with a URL, a non-executable file, etc.,
      the behavior is implementation-defined.

      Returns something false (0 or None) on success; returns something
      True (e.g., an error code from open or xdg-open) or throws on failure.
      However, note that in some cases the command may succeed without
      actually launching the targeted program."""
    if sys.platform == 'darwin':
        ret = subprocess.call(['open', program])
    elif sys.platform.startswith('win'):
        ret = os.startfile(os.path.normpath(program))
    else:
        ret = subprocess.call(['xdg-open', program])
    return ret



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'displayFrameUI.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(954, 442)
        self.layout = QtWidgets.QHBoxLayout(Form)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout.setObjectName("layout")
        self.frame = QtWidgets.QFrame(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setMinimumSize(QtCore.QSize(40, 0))
        self.frame.setMaximumSize(QtCore.QSize(40, 16777215))
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_2.setContentsMargins(4, 0, 4, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.pushLeft = QtWidgets.QPushButton(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushLeft.sizePolicy().hasHeightForWidth())
        self.pushLeft.setSizePolicy(sizePolicy)
        self.pushLeft.setMinimumSize(QtCore.QSize(30, 0))
        self.pushLeft.setMaximumSize(QtCore.QSize(30, 16777215))
        self.pushLeft.setObjectName("pushLeft")
        self.verticalLayout_2.addWidget(self.pushLeft)
        self.pushLeftLast = QtWidgets.QPushButton(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushLeftLast.sizePolicy().hasHeightForWidth())
        self.pushLeftLast.setSizePolicy(sizePolicy)
        self.pushLeftLast.setMinimumSize(QtCore.QSize(30, 0))
        self.pushLeftLast.setMaximumSize(QtCore.QSize(30, 16777215))
        self.pushLeftLast.setObjectName("pushLeftLast")
        self.verticalLayout_2.addWidget(self.pushLeftLast)
        self.layout.addWidget(self.frame)
        self.splitter = QtWidgets.QSplitter(Form)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.imageWindow = QtWidgets.QFrame(self.splitter)
        self.imageWindow.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.imageWindow.setFrameShadow(QtWidgets.QFrame.Raised)
        self.imageWindow.setLineWidth(3)
        self.imageWindow.setObjectName("imageWindow")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.imageWindow)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.labelCurrent = QtWidgets.QLabel(self.imageWindow)
        self.labelCurrent.setMinimumSize(QtCore.QSize(0, 20))
        self.labelCurrent.setMaximumSize(QtCore.QSize(16777215, 20))
        self.labelCurrent.setAlignment(QtCore.Qt.AlignCenter)
        self.labelCurrent.setObjectName("labelCurrent")
        self.verticalLayout_3.addWidget(self.labelCurrent)
        self.imageFrame = QtWidgets.QFrame(self.imageWindow)
        self.imageFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.imageFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.imageFrame.setObjectName("imageFrame")
        self.verticalLayout_3.addWidget(self.imageFrame)
        self.imageToolbar = QtWidgets.QFrame(self.imageWindow)
        self.imageToolbar.setMaximumSize(QtCore.QSize(16777215, 40))
        self.imageToolbar.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.imageToolbar.setFrameShadow(QtWidgets.QFrame.Raised)
        self.imageToolbar.setObjectName("imageToolbar")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.imageToolbar)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.imageIntRaw = QtWidgets.QComboBox(self.imageToolbar)
        self.imageIntRaw.setObjectName("imageIntRaw")
        self.imageIntRaw.addItem("")
        self.imageIntRaw.addItem("")
        self.horizontalLayout_2.addWidget(self.imageIntRaw)
        self.imageMethod = QtWidgets.QComboBox(self.imageToolbar)
        self.imageMethod.setObjectName("imageMethod")
        self.imageMethod.addItem("")
        self.imageMethod.addItem("")
        self.horizontalLayout_2.addWidget(self.imageMethod)
        self.imageUnit = QtWidgets.QComboBox(self.imageToolbar)
        self.imageUnit.setObjectName("imageUnit")
        self.imageUnit.addItem("")
        self.imageUnit.addItem("")
        self.horizontalLayout_2.addWidget(self.imageUnit)
        self.imageNRP = QtWidgets.QComboBox(self.imageToolbar)
        self.imageNRP.setObjectName("imageNRP")
        self.imageNRP.addItem("")
        self.imageNRP.addItem("")
        self.imageNRP.addItem("")
        self.horizontalLayout_2.addWidget(self.imageNRP)
        self.imageMask = QtWidgets.QCheckBox(self.imageToolbar)
        self.imageMask.setObjectName("imageMask")
        self.horizontalLayout_2.addWidget(self.imageMask)
        self.setMaskButton = QtWidgets.QPushButton(self.imageToolbar)
        self.setMaskButton.setObjectName("setMaskButton")
        self.horizontalLayout_2.addWidget(self.setMaskButton)
        self.shareAxis = QtWidgets.QCheckBox(self.imageToolbar)
        self.shareAxis.setObjectName("shareAxis")
        self.horizontalLayout_2.addWidget(self.shareAxis)
        self.verticalLayout_3.addWidget(self.imageToolbar)
        self.plotWindow = QtWidgets.QFrame(self.splitter)
        self.plotWindow.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.plotWindow.setFrameShadow(QtWidgets.QFrame.Raised)
        self.plotWindow.setLineWidth(3)
        self.plotWindow.setObjectName("plotWindow")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.plotWindow)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.plotFrame = QtWidgets.QFrame(self.plotWindow)
        self.plotFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.plotFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.plotFrame.setObjectName("plotFrame")
        self.verticalLayout_4.addWidget(self.plotFrame)
        self.plotToolBar = QtWidgets.QFrame(self.plotWindow)
        self.plotToolBar.setMaximumSize(QtCore.QSize(16777215, 40))
        self.plotToolBar.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.plotToolBar.setFrameShadow(QtWidgets.QFrame.Raised)
        self.plotToolBar.setObjectName("plotToolBar")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.plotToolBar)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.plotMethod = QtWidgets.QComboBox(self.plotToolBar)
        self.plotMethod.setObjectName("plotMethod")
        self.plotMethod.addItem("")
        self.plotMethod.addItem("")
        self.horizontalLayout.addWidget(self.plotMethod)
        self.plotUnit = QtWidgets.QComboBox(self.plotToolBar)
        self.plotUnit.setObjectName("plotUnit")
        self.plotUnit.addItem("")
        self.plotUnit.addItem("")
        self.horizontalLayout.addWidget(self.plotUnit)
        self.plotNRP = QtWidgets.QComboBox(self.plotToolBar)
        self.plotNRP.setObjectName("plotNRP")
        self.plotNRP.addItem("")
        self.plotNRP.addItem("")
        self.plotNRP.addItem("")
        self.horizontalLayout.addWidget(self.plotNRP)
        self.plotOverlay = QtWidgets.QCheckBox(self.plotToolBar)
        self.plotOverlay.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.plotOverlay.setObjectName("plotOverlay")
        self.horizontalLayout.addWidget(self.plotOverlay)
        self.verticalLayout_4.addWidget(self.plotToolBar)
        self.layout.addWidget(self.splitter)
        self.frame_2 = QtWidgets.QFrame(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_2.sizePolicy().hasHeightForWidth())
        self.frame_2.setSizePolicy(sizePolicy)
        self.frame_2.setMinimumSize(QtCore.QSize(40, 0))
        self.frame_2.setMaximumSize(QtCore.QSize(40, 16777215))
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout.setContentsMargins(4, 0, 4, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.pushRight = QtWidgets.QPushButton(self.frame_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushRight.sizePolicy().hasHeightForWidth())
        self.pushRight.setSizePolicy(sizePolicy)
        self.pushRight.setMinimumSize(QtCore.QSize(30, 0))
        self.pushRight.setMaximumSize(QtCore.QSize(30, 16777215))
        self.pushRight.setObjectName("pushRight")
        self.verticalLayout.addWidget(self.pushRight)
        self.pushRightLast = QtWidgets.QPushButton(self.frame_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushRightLast.sizePolicy().hasHeightForWidth())
        self.pushRightLast.setSizePolicy(sizePolicy)
        self.pushRightLast.setMinimumSize(QtCore.QSize(30, 0))
        self.pushRightLast.setMaximumSize(QtCore.QSize(30, 16777215))
        self.pushRightLast.setObjectName("pushRightLast")
        self.verticalLayout.addWidget(self.pushRightLast)
        self.layout.addWidget(self.frame_2)

        self.retranslateUi(Form)
        self.imageMethod.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.pushLeft.setText(_translate("Form", "<"))
        self.pushLeftLast.setText(_translate("Form", "<<"))
        self.labelCurrent.setText(_translate("Form", "Current"))
        self.imageIntRaw.setItemText(0, _translate("Form", "Integrated"))
        self.imageIntRaw.setItemText(1, _translate("Form", "Raw"))
        self.imageMethod.setItemText(0, _translate("Form", "Multi. Geo."))
        self.imageMethod.setItemText(1, _translate("Form", "By image"))
        self.imageUnit.setItemText(0, _translate("Form", "2 u038"))
        self.imageUnit.setItemText(1, _translate("Form", "q (A-1)"))
        self.imageNRP.setItemText(0, _translate("Form", "Normalized"))
        self.imageNRP.setItemText(1, _translate("Form", "Raw"))
        self.imageNRP.setItemText(2, _translate("Form", "Pixel Count"))
        self.imageMask.setText(_translate("Form", "Mask"))
        self.setMaskButton.setText(_translate("Form", "Set mask..."))
        self.shareAxis.setText(_translate("Form", "Share Axis"))
        self.plotMethod.setItemText(0, _translate("Form", "Multi. Geo."))
        self.plotMethod.setItemText(1, _translate("Form", "By image"))
        self.plotUnit.setItemText(0, _translate("Form", "2 u\\u03B8"))
        self.plotUnit.setItemText(1, _translate("Form", "q (A-1)"))
        self.plotNRP.setItemText(0, _translate("Form", "Normalized"))
        self.plotNRP.setItemText(1, _translate("Form", "Raw"))
        self.plotNRP.setItemText(2, _translate("Form", "Pixel Count"))
        self.plotOverlay.setText(_translate("Form", "Overlay"))
        self.pushRight.setText(_translate("Form", ">"))
        self.pushRightLast.setText(_translate("Form", ">>"))



import os, imp, time
from fnmatch import filter

from multiprocessing import Queue, Process

class Watcher(Process):
    """
    Run background process to monitor when new files are added to watched folders

    Inheritance:
        Process: from multiprocessing module

    Attributes:
        watchpPaths: list of folders to watch
        filetypes: list of filetypes to watch in each folder
        pollingPeriod: interval between polling the folders
        queues: Dictionary of queues in which changes in each folder are stored
        command_q: Queue object from multuprocessing module
        verbose: Flag to enable output messages (for debugging)
    """
    def __init__(self,
                 watchPaths = ["Z:\\", "P:\\bl2-1"],
                 filetypes = ['raw', 'pdi'],
                 pollingPeriod = 0.1,
                 queues = {},
                 command_q = Queue(),
                 verbose=False,
                 **kwargs):
        
        super().__init__()
        
        self.watchPaths = watchPaths
        self.filetypes = filetypes
        self.pollingPeriod = pollingPeriod
        if len(queues) < 1:
            self.queues = {ftype: Queue() for ftype in filetypes}
        else:
            self.queues = queues
        self.command_q = command_q
        self.verbose = verbose
        
    
    def run(self):
        """Run the watcher process to watch changes in folders
        """
        paths, filetypes = self.watchPaths, self.filetypes
        paths_dict = {filetype:path for filetype, path in zip(filetypes, paths)}
        
        before = dict( [filetype, dict([(f, None) for f in filter(os.listdir (path), f'*{filetype}')])]
                      for (path, filetype) in zip(paths, filetypes) )
        
        while 1:
            if not self.command_q.empty():
                command_q = self.command_q.get()
                if command_q == 'stop':
                    for _, q in self.queues.items():
                        q.put('BREAK')
                    break
                
            after = dict ( [filetype, dict ([(f, None) for f in filter(os.listdir (path), f'*{filetype}')])]
                          for (path, filetype) in zip(paths, filetypes) )
    
            for filetype in after.keys():
                added = sorted([os.path.join(paths_dict[filetype], f)
                                for f in after[filetype] 
                                if not f in before[filetype]])
                
                if added:
                    if self.verbose: print(f"{filetype} - Added: {added}")
                    for f in added:
                        self.queues[filetype].put(f)
                        
                    before[filetype] = after[filetype]
                
            time.sleep (self.pollingPeriod)


# -*- coding: utf-8 -*-
"""
@author: thampy
"""

# Standard library imports
import os
import time
import copy

# Other imports
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from pathlib import Path

# Qt imports
from PyQt5.QtCore import Qt as pyQt
import pyqtgraph as pg
from pyqtgraph import Qt
from pyqtgraph.Qt import QtWidgets
import pyqtgraph.exporters
from pyqtgraph import ROI

# This module imports
from .ui.displayFrameUI import Ui_Form
from ...gui_utils import RectViewBox, get_rect
import xdart.utils as ut
from ...widgets import pgImageWidget, pmeshImageWidget
from xdart.utils import split_file_name

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

QFileDialog = QtWidgets.QFileDialog
QInputDialog = QtWidgets.QInputDialog
QCombo = QtWidgets.QComboBox
QDialog = QtWidgets.QDialog
_translate = Qt.QtCore.QCoreApplication.translate

formats = [
    str(f.data(), encoding='utf-8').lower() for f in
    Qt.QtGui.QImageReader.supportedImageFormats()
]

AA_inv = u'\u212B\u207B\u00B9'
Th = u'\u03B8'
Chi = u'\u03C7'
Deg = u'\u00B0'
Qxy = '<math>Q<sub>xy</sub></math>'
Qz = '<math>Q<sub>z</sub></math>'

plotUnits = [f"Q ({AA_inv})", f"2{Th} ({Deg})", f"{Chi} ({Deg})", f"Qxy ({AA_inv})", f"Qz ({AA_inv})"]
imageUnits = [f"Q-{Chi}", f"2{Th}-{Chi}", f"Qxy-Qz"]

x_labels_1D = ('Q', f"2{Th}", Chi, f'{Qxy}', f'{Qz}')
x_units_1D = (AA_inv, Deg, Deg, AA_inv, AA_inv)

x_labels_2D = ('Q', f"2{Th}", Qxy)
x_units_2D = (AA_inv, Deg, AA_inv)

y_labels_2D = (Chi, Chi, Qz)
y_units_2D = (Deg, Deg, AA_inv)

# Switch to using white background and black foreground
pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')


class displayFrameWidget(Qt.QtWidgets.QWidget):
    """Widget for displaying 2D image data and 1D plots from EwaldSphere
    objects.

    attributes:
        curve1: pyqtgraph pen, overall data line
        curve2: pyqtgraph pen, individual arch data line
        histogram: pyqtgraph HistogramLUTWidget, used for adjusting min
            and max level for image
        image: pyqtgraph ImageItem, displays the 2D data
        image_plot: pyqtgraph plot, for 2D data
        image_win: pyqtgraph GraphicsLayoutWidget, layout for the 2D
            data
        imageViewBox: RectViewBox, used to set behavior of the image
            plot
        plot: pyqtgraph plot, for 1D data
        plot_layout: QVBoxLayout, for holding the 1D plotting widgets
        plot_win: pyqtgraph GraphicsLayoutWidget, layout for the 1D
            data
        sphere: EwaldSphere, unused.
        arch: EwaldArch, currently loaded arch object
        arch_ids: List of EwaldArch indices currently loaded
        arches: Dictionary of currently loaded EwaldArches
        data_1d: Dictionary object holding all 1D data in memory
        data_2d: Dictionary object holding all 2D data in memory
        ui: Ui_Form from qtdesigner

    methods:
        get_arches_map_raw: Gets averaged 2D raw data from arches
        get_sphere_map_raw: Gets averaged (and normalized) 2D raw data for all images
        get_arches_int_2d: Gets averaged 2D rebinned data from arches
        get_sphere_int_2d: Gets overall 2D data for the sphere
        update: Updates the displayed image and plot
        update_image: Updates image data based on selections
        update_plot: Updates plot data based on selections
    """

    def __init__(self, sphere, arch, arch_ids, arches, data_1d, data_2d, parent=None):
        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)

        self.ui.slice.setText(Chi)

        # Plotting parameters
        self.ui.cmap.clear()
        self.ui.cmap.addItems(['Default'] + plt.colormaps())
        self.ui.cmap.setCurrentIndex(0)
        self.cmap = self.ui.cmap.currentText()
        self.plotMethod = self.ui.plotMethod.currentText()
        self.scale = self.ui.scale.currentText()
        self.wf_yaxis = 'Frame #'
        self.wf_start = 0
        self.wf_step = 1

        # Data object initialization
        self.sphere = sphere
        self.arch = arch
        self.arch_ids = arch_ids
        self.arches = arches
        # self.idxs = sorted(list(self.arches.keys()))
        self.arch_names = []
        self.data_1d = data_1d
        self.data_2d = data_2d
        self.bkg_1d = 0.
        self.bkg_2d = 0.
        self.bkg_map_raw = 0.

        # Get Indices of selected arches
        self.idxs = []
        self.idxs_1d = []
        self.idxs_2d = []
        self.overall = False
        self.get_idxs()

        # Plotting Variables
        self.normChannel = None
        self.overlay = None

        # Image and Binned 2D Data
        self.image_data = (None, None)
        self.binned_data = (None, None)
        self.plot_data = [np.zeros(0), np.zeros(0)]
        self.plot_data_range = [[0, 0], [0, 0]]

        # Image pane setup
        self.image_layout = Qt.QtWidgets.QHBoxLayout(self.ui.imageFrame)
        self.image_layout.setContentsMargins(0, 0, 0, 0)
        self.image_layout.setSpacing(0)
        self.image_widget = pgImageWidget(lockAspect=1, raw=True)
        self.image_layout.addWidget(self.image_widget)

        # Regrouped Image pane setup
        self.binned_layout = Qt.QtWidgets.QHBoxLayout(self.ui.binnedFrame)
        self.binned_layout.setContentsMargins(0, 0, 0, 0)
        self.binned_layout.setSpacing(0)
        self.binned_widget = pgImageWidget()
        self.binned_layout.addWidget(self.binned_widget)

        # 1D/Waterfall Plot pane setup
        self.plot_layout = Qt.QtWidgets.QHBoxLayout(self.ui.plotFrame)
        self.plot_layout.setContentsMargins(0, 0, 0, 0)
        self.plot_layout.setSpacing(0)

        # 1D Plot pane setup
        self.plot_win = pg.GraphicsLayoutWidget()
        self.plot_layout.addWidget(self.plot_win)
        self.plot_viewBox = RectViewBox()
        self.plot = self.plot_win.addPlot(viewBox=self.plot_viewBox)
        self.curves = []
        self.legend = self.plot.addLegend()
        self.legend.setFlag(self.legend.ItemIgnoresTransformations, False)
        # self.legend.setLabelTextSize('10pt')

        self.pos_label = pg.LabelItem(justify='right')
        self.plot_win.addItem(self.pos_label)
        self.pos_label.anchor(itemPos=(1, 0), parentPos=(1, 0), offset=(-20, 10))
        self.pos_label.setFixedWidth(1)
        self.trackMouse()

        # WF Plot pane setup
        self.wf_widget = pgImageWidget()
        # self.wf_widget = pmeshImageWidget()
        self.setup_wf_widget()
        self.plot_layout.addWidget(self.wf_widget)

        # Waterfall Plot setup
        if self.plotMethod == 'Waterfall':
            self.plot_win.setParent(None)
            self.plot_layout.addWidget(self.wf_widget)
        else:
            self.wf_widget.setParent(None)
            self.plot_layout.addWidget(self.plot_win)

        # All Windows Signal connections
        self.ui.normChannel.activated.connect(self.normUpdate)
        self.ui.setBkg.clicked.connect(self.setBkg)
        self.ui.scale.currentIndexChanged.connect(self.update_views)
        self.ui.cmap.currentIndexChanged.connect(self.update_views)
        self.ui.shareAxis.stateChanged.connect(self.update)
        self.ui.update2D.stateChanged.connect(self.enable_2D_buttons)

        # 2D Window Signal connections
        self.ui.imageUnit.activated.connect(self.update_binned)
        self.ui.imageUnit.activated.connect(self._update_slice_range)
        # self.ui.imageMask.stateChanged.connect(self.update_binned)

        # 1D Window Signal connections
        self.ui.plotMethod.currentIndexChanged.connect(self.update_plot_view)
        self.ui.yOffset.valueChanged.connect(self.update_plot_view)
        self.ui.plotUnit.activated.connect(self._set_slice_range)
        self.ui.plotUnit.activated.connect(self.update_plot)
        # self.ui.showLegend.stateChanged.connect(self.update_plot_view)
        self.ui.showLegend.stateChanged.connect(self.update_legend)
        self.ui.slice.stateChanged.connect(self.update_plot)
        self.ui.slice.stateChanged.connect(self._update_slice_range)
        self.ui.slice_center.valueChanged.connect(self.update_plot_range)
        self.ui.slice_width.valueChanged.connect(self.update_plot_range)
        self.ui.wf_options.clicked.connect(self.popup_wf_options)

        # Clear 1D Plot Buttons
        self.ui.clear_1D.clicked.connect(self.clear_1D)

        # Save Image/Data Buttons
        self.ui.save_2D.clicked.connect(self.save_image)
        self.ui.save_1D.clicked.connect(self.save_1D)

        # Initialize image units
        self.set_axes()
        self._set_slice_range(initialize=True)

        # Setup Waterfall Options Popup
        self.wf_dialog = QDialog()
        self.wf_yaxis_widget = QCombo()
        self.wf_start_widget = QtWidgets.QDoubleSpinBox()
        self.wf_step_widget = QtWidgets.QDoubleSpinBox()
        self.wf_accept_button = QtWidgets.QPushButton('Okay')
        self.wf_cancel_button = QtWidgets.QPushButton('Cancel')

    def get_idxs(self):
        """ Return selected arch indices
        """
        # ic()
        self.idxs, self.idxs_1d, self.idxs_2d = [], [], []
        if len(self.arch_ids) == 0 or self.arch_ids[0] == 'No data':
            return

        # with self.sphere.sphere_lock:
        # if 'Overall' in self.arch_ids:
        if len(self.arch_ids) == len(self.sphere.arches.index) > 1:
            self.overall = True
            self.idxs = sorted(np.asarray(self.sphere.arches.index, dtype=int))
        else:
            self.overall = False
            self.idxs = sorted(self.arch_ids)

        try:
            self.idxs = list(np.asarray(self.idxs, dtype=int))
        except ValueError:
            return
        self.idxs_1d = [int(idx) for idx in self.idxs if idx in self.data_1d.keys()]
        self.idxs_2d = [int(idx) for idx in self.idxs if idx in self.data_2d.keys()]

    def update_plot_range(self):
        if self.ui.slice.isChecked():
            self.update_plot()

    def setup_1d_layout(self):
        """Setup the layout for 1D plot
        """
        self.wf_widget.setParent(None)
        self.plot_layout.addWidget(self.plot_win)

        self.ui.wf_options.setEnabled(False)
        if len(self.plot_data[1]) > 1:
            self.ui.wf_options.setEnabled(True)
            self.wf_yaxis_widget.setEnabled(False)

    def setup_wf_widget(self):
        self.plot_layout.addWidget(self.wf_widget)

        # Waterfall Plot setup
        if self.plotMethod == 'Waterfall':
            self.plot_win.setParent(None)
            self.plot_layout.addWidget(self.wf_widget)
        else:
            self.wf_widget.setParent(None)
            self.plot_layout.addWidget(self.plot_win)

    def setup_wf_layout(self):
        """Setup the layout for WF plot
        """
        self.plot_win.setParent(None)
        self.plot_layout.addWidget(self.wf_widget)

        self.ui.wf_options.setEnabled(True)
        self.wf_yaxis_widget.setEnabled(True)

    def popup_wf_options(self):
        """
        Popup Qt Window to select options for Waterfall Plot
        Options include Y-axis unit and number of points to skip
        """
        if self.wf_dialog.layout() is None:
            self.setup_wf_options_widget()

        self.wf_dialog.show()

    def setup_wf_options_widget(self):
        """
        Setup y-axis option for Waterfall plot
        Setup first image and step size for wf and overlay plots
        """
        layout = QtWidgets.QGridLayout()
        self.wf_dialog.setLayout(layout)

        layout.addWidget(QtWidgets.QLabel('Y-Axis'), 0, 0)
        layout.addWidget(QtWidgets.QLabel('Start'), 0, 1)
        layout.addWidget(QtWidgets.QLabel('Step'), 0, 2)

        layout.addWidget(self.wf_yaxis_widget, 1, 0)
        layout.addWidget(self.wf_start_widget, 1, 1)
        layout.addWidget(self.wf_step_widget, 1, 2)

        layout.addWidget(self.wf_accept_button, 2, 1)
        layout.addWidget(self.wf_cancel_button, 2, 2)

        arch = self.data_1d[self.idxs_1d[0]]
        counters = list(arch.scan_info.keys())
        counters = ['Frame #', 'Time (s)', 'Time (minutes)'] + counters
        self.wf_yaxis_widget.addItems(counters)

        self.wf_start_widget.setDecimals(0)
        self.wf_start_widget.setRange(1, 1000)

        self.wf_step_widget.setDecimals(0)
        self.wf_step_widget.setRange(1, 100)

        self.wf_accept_button.clicked.connect(self.get_wf_option)
        self.wf_cancel_button.clicked.connect(self.close_wf_popup)

    def get_wf_option(self):
        self.wf_yaxis = self.wf_yaxis_widget.currentText()

        self.wf_start = int(self.wf_start_widget.value()) - 1
        self.wf_step = int(self.wf_step_widget.value())

        self.close_wf_popup()
        self.update_plot_view()

    def close_wf_popup(self):
        self.wf_dialog.close()

    def enable_2D_buttons(self):
        """Disable buttons if update 2D is unchecked"""
        pass

    def set_axes(self):
        """Disable/Enable Qz-Qxy option if we are/are not in GI mode"""

        self.ui.imageUnit.setItemText(0, _translate("Form", imageUnits[0]))
        self.ui.imageUnit.setItemText(1, _translate("Form", imageUnits[1]))

        self.ui.plotUnit.setItemText(0, _translate("Form", plotUnits[0]))
        self.ui.plotUnit.setItemText(1, _translate("Form", plotUnits[1]))
        self.ui.plotUnit.setItemText(2, _translate("Form", plotUnits[2]))

        if self.sphere.gi:
            if self.ui.imageUnit.count() != 3:
                self.ui.imageUnit.addItem(_translate("Form", imageUnits[2]))
            if self.ui.plotUnit.count() != 5:
                self.ui.plotUnit.addItem(_translate("Form", plotUnits[3]))
                self.ui.plotUnit.addItem(_translate("Form", plotUnits[4]))

    def _updated(self):
        """Check if there is data to update
        """
        if (len(self.arch_ids) == 0) or (self.sphere.name == 'null_main'):
            return False
        if (len(self.data_1d) == 0) or (len(self.idxs_1d) == 0):
            return False
        # if self.ui.update2D.isChecked() and (len(self.idxs_2d) == 0) and (not self.overall):
        #     return False

        return True

    def update(self):
        """Updates image and plot frames based on toolbar options
        """
        # ic()
        self.get_idxs()

        if not self._updated():
            return True

        if self.ui.shareAxis.isChecked() and (self.ui.imageUnit.currentIndex() < 2):
            self.ui.plotUnit.setCurrentIndex(self.ui.imageUnit.currentIndex())
            self.ui.plotUnit.setEnabled(False)
            self.plot.setXLink(self.binned_widget.image_plot)
        else:
            self.plot.setXLink(None)
            self.ui.plotUnit.setEnabled(True)

        try:
            self.update_plot()
        except TypeError:
            return False

        if self.ui.update2D.isChecked():
            try:
                self.update_image()
            except TypeError:
                return False
            try:
                self.update_binned()
            except TypeError:
                return False

        # Apply label to 2D view
        if self.ui.update2D.isChecked():
            self.update_2d_label()

        return True

    def update_views(self):
        """Updates 2D (if flag is selected) and 1D views
        """
        if not self._updated():
            return True

        self.cmap = self.ui.cmap.currentText()
        self.plotMethod = self.ui.plotMethod.currentText()
        self.scale = self.ui.scale.currentText()

        if self.ui.update2D.isChecked():
            self.update_image_view()
            self.update_binned_view()
            self.update_2d_label()
        self.update_plot_view()

    def update_image(self):
        """Updates image plotted in image frame
        """
        mask = None
        if self.overall and len(self.arch_ids) > 1:
            data = self.get_sphere_map_raw()
        else:
            data = self.get_arches_map_raw()
            if data is None:
                return

            # Apply Mask
            arch_2d = self.data_2d[self.idxs_2d[0]]
            mask = arch_2d['mask']
        data = np.asarray(data, dtype=float)

        # Apply Mask
        global_mask = self.sphere.global_mask if self.sphere.global_mask is not None else []
        mask = mask if mask is not None else []
        mask = np.asarray(np.unique(np.append(mask, global_mask)), dtype=int)
        if len(mask) > 0:
            mask = np.unravel_index(mask, data.shape)
            data[mask] = np.nan

        # Subtract background
        data -= self.bkg_map_raw

        data = data.T[:, ::-1]

        # Get Bounding Rectangle
        rect = get_rect(np.arange(data.shape[0]), np.arange(data.shape[1]))

        self.image_data = (data, rect)
        self.update_image_view()

    def update_image_view(self):
        data, rect = self.image_data

        # self.image_widget.setImage(data.T[:, ::-1], scale=self.scale, cmap=self.cmap)
        self.image_widget.setImage(data, scale=self.scale, cmap=self.cmap)
        self.image_widget.setRect(rect)

        self.image_widget.image_plot.setLabel("bottom", 'x (Pixels)')
        self.image_widget.image_plot.setLabel("left", 'y (Pixels)')

    def update_binned(self):
        """Updates image plotted in image frame
        """
        if self.ui.shareAxis.isChecked() and (self.ui.imageUnit.currentIndex() < 2):
            self.ui.plotUnit.setCurrentIndex(self.ui.imageUnit.currentIndex())
            self.update_plot()

        # if 'Overall' in self.arch_ids:
        if self.overall and len(self.arch_ids) > 1:
            intensity, xdata, ydata = self.get_sphere_int_2d()
        else:
            intensity, xdata, ydata = self.get_arches_int_2d()

        if intensity is None:
            return

        # Subtract background
        intensity -= self.bkg_2d

        rect = get_rect(xdata, ydata)
        self.binned_data = (intensity, rect)
        self.update_binned_view()

        return

    def update_binned_view(self):
        data, rect = self.binned_data

        self.binned_widget.setImage(data.T[:, ::-1], scale=self.scale, cmap=self.cmap)
        self.binned_widget.setRect(rect)

        imageUnit = self.ui.imageUnit.currentIndex()
        self.binned_widget.image_plot.setLabel(
            "bottom", x_labels_2D[imageUnit], units=x_units_2D[imageUnit]
        )
        self.binned_widget.image_plot.setLabel(
            "left", y_labels_2D[imageUnit], units=y_units_2D[imageUnit]
        )

        self.show_slice_overlay()
        return data

    def update_2d_label(self):
        """Updates 2D Label
        """
        # Sets title text
        label = self.sphere.name
        if len(label) > 40:
            label = f'{label[:10]}.....{label[-30:]}'

        if (self.overall or self.sphere.single_img) and (len(self.arch_ids) > 1):
            self.ui.labelCurrent.setText(label)
        elif self.sphere.series_average:
            self.ui.labelCurrent.setText(label)
        elif len(self.arch_ids) > 1:
            self.ui.labelCurrent.setText(f'{label} [Average]')
        else:
            self.ui.labelCurrent.setText(f'{label}_{self.arch_ids[0]}')

    def update_plot(self):
        """Updates data in plot frame
        """
        if (self.sphere.name == 'null_main') or (len(self.arch_ids) == 0):
            data = (np.arange(100), np.arange(100))
            return data

        # Get 1D data for all arches
        ydata, xdata = self.get_arches_int_1d()
        if self.sphere.series_average:
            arch_names = [self.sphere.name]
        else:
            arch_names = [f'{self.sphere.name}_{i}' for i in self.idxs]

        # Subtract background
        ydata -= self.bkg_1d
        if ydata.ndim == 1:
            ydata = ydata[np.newaxis, :]

        if ((self.plot_data[0].shape == xdata.shape) and
                (self.plotMethod in ['Overlay', 'Waterfall'])):
            for (arch_name, data) in zip(arch_names, ydata):
                if arch_name not in self.arch_names:
                    self.plot_data[1] = np.vstack((self.plot_data[1], data))
                    self.arch_names.append(arch_name)
        else:
            self.plot_data = [xdata, ydata]
            self.arch_names.clear()
            self.arch_names = arch_names

        xdata, ydata = self.plot_data
        self.plot_data_range = [[xdata.min(), xdata.max()], [ydata.min(), ydata.max()]]
        self.update_plot_view()

        # self.save_1D(auto=True)

    def update_plot_view(self):
        """Updates 1D view of data in plot frame
        """
        if (len(self.arch_ids) == 0) or len(self.data_1d) == 0:
            return

        # Clear curves
        [curve.clear() for curve in self.curves]
        self.curves.clear()

        self.plotMethod = self.ui.plotMethod.currentText()
        self.ui.yOffset.setEnabled(False)
        if (self.plotMethod in ['Overlay', 'Single']) and (len(self.arch_names) > 1):
            self.ui.yOffset.setEnabled(True)

        # Only switch to WF plot if more than three curves. Definitely switch if more than 15!
        if (self.plotMethod == 'Waterfall' and len(self.plot_data[1]) > 3) or len(self.plot_data[1]) > 15:
            self.update_wf()
        else:
            self.update_1d_view()

    def update_1d_view(self):
        """Updates data in 1D plot Frame
        """
        self.setup_1d_layout()
        xdata_, ydata_ = self.plot_data
        s_xdata, ydata = xdata_.copy(), ydata_.copy()

        int_label = 'I'
        if self.normChannel:
            int_label = f'I / {self.normChannel}'

        self.plot.getAxis("left").setLogMode(False)
        self.plot.getAxis("bottom").setLogMode(False)
        ylabel = f'{int_label} (a.u.)'
        if self.scale == 'Log':
            if ydata.min() < 1:
                ydata -= (ydata.min() - 1.)
            ydata = np.log10(ydata)
            self.plot.getAxis("left").setLogMode(True)
            ylabel = f'Log {int_label}(a.u.)'
        elif self.scale == 'Log-Log':
            if ydata.min() < 1:
                ydata -= (ydata.min() - 1.)
            ydata = np.log10(ydata)
            self.plot.getAxis("left").setLogMode(True)
            ylabel = f'Log {int_label}(a.u.)'

            s_xdata = np.log10(s_xdata)
            self.plot.getAxis("bottom").setLogMode(True)
        elif self.scale == 'Sqrt':
            if ydata.min() < 0.:
                ydata_ = np.sqrt(np.abs(ydata))
                ydata_[ydata < 0] *= -1
                ydata = ydata_
            else:
                ydata = np.sqrt(ydata)
            ylabel = f'<math>&radic;</math>{int_label} (a.u.)'

        if ((self.plotMethod in ['Overlay', 'Waterfall']) or
                ((self.plotMethod == 'Single') and (ydata.shape[0] > 1))):
            ydata = ydata[self.wf_start::self.wf_step]
            self.setup_curves()

            offset = self.ui.yOffset.value()
            y_offset = offset / 100 * (self.plot_data_range[1][1] - self.plot_data_range[1][0])
            for nn, (curve, s_ydata) in enumerate(zip(self.curves, ydata)):
                curve.setData(s_xdata, s_ydata + y_offset*nn)

        else:
            self.setup_curves()
            s_ydata = ydata
            if self.plotMethod == 'Average':
                s_ydata = np.nanmean(s_ydata, 0)
            elif self.plotMethod == 'Sum':
                s_ydata = np.nansum(s_ydata, 0)

            self.curves[0].setData(s_xdata, s_ydata.squeeze())

        # Apply labels to plot
        plotUnit = self.ui.plotUnit.currentIndex()
        self.plot.setLabel("bottom", x_labels_1D[plotUnit], units=x_units_1D[plotUnit])
        self.plot.setLabel("left", ylabel)

        return s_xdata, s_ydata

    def update_wf(self):
        """Updates data in 1D plot Frame
        """
        self.setup_wf_layout()

        xdata_, data_ = self.plot_data
        s_xdata, data = xdata_.copy(), data_.copy()
        data = data[self.wf_start::self.wf_step, :]

        # Set YAxis Unit
        if self.wf_yaxis == 'Frame #':
            s_ydata = np.asarray(np.arange(data.shape[0]) + self.wf_start + 1, dtype=float)
        else:
            # TODO: Fix below for more WF options
            try:
                if self.wf_yaxis == 'Time (s)':
                    s_ydata = np.asarray([self.data_1d[idx].scan_info['epoch'] for idx in self.idxs])
                    s_ydata -= s_ydata.min()
                elif self.wf_yaxis == 'Time (minutes)':
                    s_ydata = np.asarray([self.data_1d[idx].scan_info['epoch'] for idx in self.idxs])/60.
                    s_ydata -= s_ydata.min()
                else:
                    s_ydata = np.asarray([self.data_1d[idx].scan_info[self.wf_yaxis] for idx in self.idxs])

                s_ydata = s_ydata[self.wf_start::self.wf_step]
            except KeyError:
                print('Counter not present in metadata')

        # rect = get_rect(s_xdata, np.arange(data.shape[0]))
        rect = get_rect(s_xdata, s_ydata)

        self.wf_widget.setImage(data.T, scale=self.scale, cmap=self.cmap)
        self.wf_widget.setRect(rect)

        plotUnit = self.ui.plotUnit.currentIndex()
        self.wf_widget.image_plot.setLabel("bottom", x_labels_1D[plotUnit],
                                           units=x_units_1D[plotUnit])
        self.wf_widget.image_plot.setLabel("left", self.wf_yaxis)

    def get_arches_map_raw(self, idxs=None):
        """Return 2D arch data for multiple arches (averaged)"""
        if idxs is None:
            # idxs = self.idxs
            idxs = self.idxs_2d

        intensity, ctr = 0., 0
        for nn, idx in enumerate(idxs):
            # arch = self.arches[int(idx)]
            arch_1d = self.data_1d[int(idx)]
            arch_2d = self.data_2d[int(idx)]
            for kk in range(3):
                try:
                    intensity += self.normalize(arch_2d['map_raw']-arch_2d['bg_raw'],
                                                arch_1d.scan_info)
                    ctr += 1
                    break
                except ValueError:
                    time.sleep(0.5)

        if ctr > 0:
            intensity /= ctr
        else:
            return None

        return np.asarray(intensity, dtype=float)

    def get_sphere_map_raw(self):
        """Returns data and QRect for data in sphere
        """
        with self.sphere.sphere_lock:
            map_raw = np.asarray(self.sphere.overall_raw, dtype=float)
            if map_raw.ndim < 2:
                self.sphere.load_from_h5(data_only=True)
                map_raw = np.asarray(self.sphere.overall_raw, dtype=float)

            norm_fac = len(self.sphere.arches.index)
            if self.normChannel:
                norm = self.sphere.scan_data[self.normChannel].sum()
                if norm > 0:
                    norm_fac = norm

            return map_raw/norm_fac

    def get_arches_int_2d(self, idxs=None):
        """Return 2D arch data for multiple arches (averaged)"""
        if idxs is None:
            idxs = self.idxs_2d

        if len(idxs) > 1:
            intensity = self.get_int_2d(self.arches['sum_int_2d'])
            xdata, ydata = self.get_xydata(self.arches['sum_int_2d'])
            return intensity, xdata, ydata

        # intensity, n_arches = 0., 0
        # for nn, idx in enumerate(idxs):
        try:
            idx = idxs[0]
        except IndexError:
            return None, None, None

        arch_1d = self.data_1d[int(idx)]
        arch_2d = self.data_2d[int(idx)]
        # ic(arch_2d['int_2d'].ttheta, arch_2d['int_2d'].i_tthChi)
        # print(arch_2d['int_2d'].ttheta, arch_2d['int_2d'].i_tthChi)
        # arch_intensity = self.get_int_2d(arch_2d['int_2d'], arch_1d)
        intensity = self.get_int_2d(arch_2d['int_2d'], arch_1d)
        # if (arch_intensity.ndim > 1) and (arch_intensity.shape[0] != 0):
        #     intensity += arch_intensity
        #     arch_xy = arch_2d
        #     n_arches += 1

        # if n_arches == 0:
        if intensity.ndim != 2:
            return None, None, None

        # intensity /= n_arches

        # xdata, ydata = self.get_xydata(arch_xy['int_2d'])
        xdata, ydata = self.get_xydata(arch_2d['int_2d'])
        return np.asarray(intensity, dtype=float), xdata, ydata

    def get_sphere_int_2d(self):
        """Returns data and QRect for data in sphere
        """
        with self.sphere.sphere_lock:
            int_2d = self.sphere.bai_2d

        intensity = self.get_int_2d(int_2d, normalize=True)

        xdata, ydata = self.get_xydata(int_2d)
        return intensity, xdata, ydata

    def get_arches_int_1d(self, idxs=None, rv='all'):
        """Return 1D data for multiple arches"""
        if idxs is None:
            idxs = self.idxs_1d

        ydata = 0.
        for nn, idx in enumerate(idxs):
            arch_1d = self.data_1d[int(idx)]
            arch_2d = self.data_2d.get(int(idx), None)
            xdata, s_ydata = self.get_int_1d(arch_1d, arch_2d, idx)
            if nn == 0:
                ydata = s_ydata
            else:
                ydata = np.vstack((ydata, s_ydata))

        if ydata.ndim == 2:
            if rv == 'average':
                ydata = np.nanmean(ydata, 0)
            elif rv == 'sum':
                ydata = np.nansum(ydata, 0)

        return ydata, xdata

    def get_int_2d(self, int_2d, arch_1d=None, normalize=True):
        """Returns the appropriate 2D data depending on the chosen axes
        I(Q, Chi) or I(Qz, Qxy)
        """
        if self.ui.imageUnit.currentIndex() == 0:
            intensity_2d = int_2d.i_qChi
        elif self.ui.imageUnit.currentIndex() == 1:
            intensity_2d = int_2d.i_tthChi
        else:
            intensity_2d = int_2d.i_QxyQz
        intensity = np.asarray(intensity_2d.copy(), dtype=float)

        if normalize:
            if arch_1d is not None:
                intensity = self.normalize(intensity, arch_1d.scan_info)
            else:
                norm_fac = len(self.sphere.arches.index)
                if self.normChannel:
                    norm = self.sphere.scan_data[self.normChannel].sum()
                    if norm > 0:
                        norm_fac = norm
                intensity /= norm_fac

        return intensity

    def get_int_1d(self, arch, arch_2d, idx):
        """Returns 1D integrated data for arch. If range is specified,
        it returns integrated data over that range
        """
        if self.ui.plotUnit.currentIndex() != 2:
            if not self.ui.slice.isChecked():
                int_1d = arch.int_1d
                if self.ui.plotUnit.currentIndex() == 3:
                    intensity = int_1d.i_qxy
                elif self.ui.plotUnit.currentIndex() == 4:
                    intensity = int_1d.i_qz
                else:
                    intensity = int_1d.norm

                ydata = self.normalize(intensity, arch.scan_info)
                xdata = self.get_xdata(int_1d)
                return xdata, ydata

        if arch_2d is None:
            return None, None

        # If a particular Chi Range is chosen...
        intensity = self.get_int_2d(arch_2d['int_2d'], arch, normalize=False)
        q, tth, chi = arch_2d['int_2d'].q, arch_2d['int_2d'].ttheta, arch_2d['int_2d'].chi

        xdata_list = [q, tth, chi]
        xdata = xdata_list[self.ui.plotUnit.currentIndex()]

        center = self.ui.slice_center.value()
        width = self.ui.slice_width.value()
        _range = [center - width, center + width]

        _inds = np.s_[:]
        if self.ui.plotUnit.currentIndex() < 2:
            slice_axis = Chi
            if self.ui.slice.isChecked():
                _inds = (_range[0] <= chi) & (chi <= _range[1])
            ydata = np.nanmean(intensity[_inds, :], axis=0)
        else:
            if self.ui.imageUnit.currentIndex() != 1:
                slice_axis = 'Q'
                if self.ui.slice.isChecked():
                    _inds = (_range[0] <= q) & (q <= _range[1])
            else:
                slice_axis = f'2{Th}'
                if self.ui.slice.isChecked():
                    _inds = (_range[0] <= tth) & (tth <= _range[1])
            ydata = np.nanmean(intensity[:, _inds], axis=1)

        self.ui.slice.setText(f'{slice_axis} Range')
        self.show_slice_overlay()

        ydata = self.normalize(ydata, arch.scan_info)
        return xdata, ydata

    def get_xydata(self, int_2d):
        """Reads the unit box and returns appropriate xdata

        args:
            box: QComboBox, list of options
            int_data: int_nd_data object, data to parse

        returns:
            xdata: numpy array, x axis data for plot.
        """
        imageUnit = self.ui.imageUnit.currentIndex()
        if imageUnit == 0:  # Q-Chi
            xdata, ydata = int_2d.q, int_2d.chi
        elif imageUnit == 1:  # 2Th-Chi
            xdata, ydata = int_2d.ttheta, int_2d.chi
        else:  # Qzx-Qz
            xdata, ydata = int_2d.qxy, int_2d.qz

        return xdata, ydata

    def get_xdata(self, int_data):
        """Reads the unit box and returns appropriate xdata

        args:
            box: QComboBox, list of options
            int_data: int_nd_data object, data to parse

        returns:
            xdata: numpy array, x axis data for plot.
        """
        unit = self.ui.plotUnit.currentIndex()
        if unit == 0:
            xdata = int_data.q
        elif unit == 1:
            xdata = int_data.ttheta
        elif unit == 2:
            xdata = self.get_chi_1d()
        elif unit == 3:
            xdata = int_data.qxy
        else:
            xdata = int_data.qz
        return xdata

    def normalize(self, int_data, scan_info):
        """Reads the norm, raw, pcount option box and returns
        appropriate ydata.

        args:
            box: QComboBox, list of choices for picking data to return
            int_data: int_nd_data object, data to parse

        returns:
            data: numpy array, non-zero region from nzarray based on
                choices in box
            corners: tuple, the bounds of the non-zero region of the
                dataset
        """
        try:
            intensity = np.asarray(int_data.copy(), dtype=float)
        except AttributeError:
            return np.zeros((10, 10))

        # normChannel = self.ui.normChannel.currentText()
        # if normChannel in scan_info.keys():
        #     self.normChannel = normChannel
        # elif normChannel.upper() in scan_info.keys():
        #     self.normChannel = normChannel.upper()
        # elif normChannel.lower() in scan_info.keys():
        #     self.normChannel = normChannel.lower()
        # else:
        #     self.normChannel = None

        normChannel = self.get_normChannel(scan_data_keys=scan_info.keys())
        if normChannel and (scan_info[normChannel] > 0):
            intensity /= scan_info[normChannel]

        return intensity

    def get_normChannel(self, scan_data_keys=None):
        """Check to see if normalization channel exists in metadata and return name"""
        normChannel = self.ui.normChannel.currentText()
        if normChannel == 'sec':
            normChannel = {'sec', 'seconds', 'Seconds', 'Sec', 'SECONDS', 'SEC'}
        elif normChannel == 'Monitor':
            normChannel = {'Monitor', 'monitor', 'mon', 'Mon', 'MON', 'MONITOR'}
        else:
            normChannel = {normChannel, normChannel.lower(), normChannel.upper()}
        if scan_data_keys is None:
            scan_data_keys = self.sphere.scan_data.columns
        normChannel = normChannel.intersection(scan_data_keys)
        return normChannel.pop() if len(normChannel) > 0 else None

    def normUpdate(self):
        """Update plots if norm channel exists"""
        self.normChannel = self.get_normChannel()
        if self.normChannel and (self.sphere.scan_data[self.normChannel].sum() == 0.):
            self.normChannel = None
        self.update()

    def setBkg(self):
        """Sets selected points as background.
        If background is already selected, it unsets it"""
        if (len(self.arch_ids) == 0) or (len(self.idxs) == 0):
            return

        if self.ui.setBkg.text() == 'Set Bkg':
            idxs = self.arch_ids
            # if 'Overall' in self.arch_ids:
            if self.overall:
                idxs = sorted(list(self.sphere.arches.index))

            self.bkg_1d, _ = self.get_arches_int_1d(idxs, rv='average')
            self.bkg_2d, _, _ = self.get_arches_int_2d(idxs)
            self.bkg_map_raw = self.get_arches_map_raw(idxs)
            if self.bkg_map_raw is None:
                self.bkg_map_raw = 0.
            self.ui.setBkg.setText('Clear Bkg')
        else:
            self.bkg_1d = 0.
            self.bkg_2d = 0.
            self.bkg_map_raw = 0.
            self.ui.setBkg.setText('Set Bkg')

        self.update()
        return

    def setup_curves(self):
        """Initialize curves for line plots
        """
        self.curves.clear()
        self.legend.clear()

        arch_ids = self.arch_names[self.wf_start::self.wf_step]
        if (self.plotMethod in ['Sum', 'Average']) and (len(self.arch_names) > 1):
            arch_ids = f'{self.plotMethod} [{self.arch_names[0]}'
            for arch_name in self.arch_names[1:]:
                arch_ids += f', {arch_name}'
            arch_ids = [arch_ids + ']']

        colors = self.get_colors()
        self.curves = [self.plot.plot(
            pen=color,
            symbolBrush=color,
            symbolPen=color,
            symbolSize=4,
            name=arch_id,
        ) for (color, arch_id) in zip(colors, arch_ids)]

        if not self.ui.showLegend.isChecked():
            self.legend.clear()

    def clear_1D(self):
        """Initialize curves for line plots
        """
        self.arch_names.clear()
        self.arch_ids.clear()
        self.plot_data = [np.zeros(0), np.zeros(0)]
        self.setup_1d_layout()
        self.plot.clear()

    def update_legend(self):
        if not self.ui.showLegend.isChecked():
            self.legend.hide()
        else:
            self.legend.show()

    def _set_slice_range(self, _=None, initialize=False):
        if self.ui.plotUnit.currentIndex() == 2:
            if self.ui.imageUnit.currentIndex() != 1:
                self.ui.slice.setText('Q Range')
                self.ui.slice_center.setRange(0, 25)
                self.ui.slice_width.setRange(0, 30)
                self.ui.slice_center.setSingleStep(0.1)
                self.ui.slice_width.setSingleStep(0.1)
                if initialize:
                    self.ui.slice_center.setValue(2)
                    self.ui.slice_width.setValue(0.5)
            else:
                self.ui.slice.setText(f'2{Th} Range')
                self.ui.slice_center.setRange(-180, 180)
                self.ui.slice_width.setRange(0, 180)
                self.ui.slice_center.setValue(10)
                self.ui.slice_width.setValue(5)
                self.ui.slice_center.setSingleStep(1)
                self.ui.slice_width.setSingleStep(1)
                if initialize:
                    self.ui.slice_center.setValue(10)
                    self.ui.slice_width.setValue(5)
        else:
            self.ui.slice.setText(f'{Chi} Range')
            self.ui.slice_center.setMinimum(-180)
            self.ui.slice_center.setMaximum(180)
            self.ui.slice_width.setMinimum(0)
            self.ui.slice_width.setMaximum(270)
            self.ui.slice_center.setSingleStep(1)
            self.ui.slice_width.setSingleStep(1)
            if initialize:
                self.ui.slice_center.setProperty("value", 0)
                self.ui.slice_width.setProperty("value", 10)

    def _update_slice_range(self):

        if not self.ui.slice.isChecked():
            self.clear_slice_overlay()
            return

        imageUnit = self.ui.imageUnit.currentIndex()
        plotUnit = self.ui.plotUnit.currentIndex()
        if (plotUnit != 2) or (imageUnit == 2):
            self.update_plot()
            return

        cen = self.ui.slice_center.value()
        wid = self.ui.slice_width.value()
        _range = np.array([cen - wid, cen + wid])
        wavelength = self.data_1d[self.idxs_1d[0]].integrator.wavelength

        if imageUnit == 0:
            if self.ui.slice.text() == f'2{Th} Range':
                self.ui.slice.setText('Q Range')
                _range = ((4 * np.pi / (wavelength * 1e10)) * np.sin(np.radians(_range / 2)))
                # cen = ((4 * np.pi / (wavelength * 1e10)) * np.sin(np.radians(cen / 2)))
                # wid = ((4 * np.pi / (wavelength * 1e10)) * np.sin(np.radians(wid / 2)))
        else:
            if self.ui.slice.text() == f'Q Range':
                self.ui.slice.setText(f'2{Th} Range')
                _range = (2 * np.degrees(np.arcsin(_range * (wavelength * 1e10) / (4 * np.pi))))
                # cen = (2 * np.degrees(np.arcsin(cen * (wavelength * 1e10) / (4 * np.pi))))
                # wid = (2 * np.degrees(np.arcsin(wid * (wavelength * 1e10) / (4 * np.pi))))

        cen = (_range[-1] + _range[0]) / 2.
        wid = (_range[-1] - _range[0]) / 2.
        self.ui.slice_center.setValue(cen)
        self.ui.slice_width.setValue(wid)
        self._set_slice_range()

        self.show_slice_overlay()

    def show_slice_overlay(self):
        """
        Shows the slice integration region on 2D Binned plot
        """
        self.clear_slice_overlay()

        if (not self.ui.slice.isChecked()) or (self.ui.imageUnit.currentIndex() == 2):
            return

        center = self.ui.slice_center.value()
        width = self.ui.slice_width.value()
        _range = [center-width, center+width]

        binned_data, rect = self.binned_data

        if self.ui.plotUnit.currentIndex() < 2:
            _range = [max(rect.top(), _range[0]), min(rect.top() + rect.height(), _range[1])]
            width = (_range[1] - _range[0])/2.
            self.overlay = ROI(
                [rect.left(), _range[0]], [rect.width(), 2*width],
                pen=(255, 255, 255),
                maxBounds=rect
            )
        else:
            _range = [max(rect.left(), _range[0]), min(rect.left() + rect.width(), _range[1])]
            width = (_range[1] - _range[0])/2.
            self.overlay = ROI(
                [_range[0], rect.top()], [2*width, rect.height()],
                pen=(255, 255, 255),
                maxBounds=rect
            )

        self.binned_widget.imageViewBox.addItem(self.overlay, ignoreBounds=True)

    def clear_slice_overlay(self):
        """Clear the overlay that shows integration slice"""
        if self.overlay is not None:
            self.binned_widget.imageViewBox.removeItem(self.overlay)
            self.overlay = None

    def save_image(self):
        """Saves currently displayed image. Formats are automatically
        grabbed from Qt. Also implements tiff saving.
        """
        ext_filter = "Images ("
        for f in formats:
            ext_filter += "*." + f + " "

        dialog = QFileDialog()
        fname, _ = dialog.getSaveFileName(
            dialog,
            filter=ext_filter,
            caption='Save as...',
            options=QFileDialog.DontUseNativeDialog
        )
        if fname == '':
            return

        # Save as image
        data, rect = self.binned_data
        scene = self.binned_widget.imageViewBox.scene()
        exporter = pyqtgraph.exporters.ImageExporter(scene)
        h = exporter.params.param('height').value()
        w = exporter.params.param('width').value()
        h_new = 2000
        w_new = int(np.round(w/h * h_new, 0))
        exporter.params.param('height').setValue(h_new)
        exporter.params.param('width').setValue(w_new)
        exporter.export(fname)

        # Save as Numpy array as well
        directory, base_name, ext = split_file_name(fname)
        save_fname = os.path.join(directory, base_name)
        np.save(f'{save_fname}.npy', data)

    def save_1D(self, auto=False):
        """Saves currently displayed data. Currently supports .xye
        and .csv.
        """
        fname = f'{self.sphere.name}'
        if not auto:
            path = QFileDialog().getExistingDirectory(
                caption='Choose Save Directory',
                directory='',
                options=(QFileDialog.ShowDirsOnly | QFileDialog.DontUseNativeDialog)
            )

            inp_dialog = QtWidgets.QInputDialog()
            suffix, ok = inp_dialog.getText(inp_dialog, 'Enter Suffix to be added to File Name', 'Suffix', text='')
            if not ok:
                return
            if suffix != '':
                fname += f'_{suffix}'
        else:
            path = os.path.dirname(self.sphere.data_file)
            path = os.path.join(path, self.sphere.name)
            Path(path).mkdir(parents=True, exist_ok=True)

        fname = os.path.join(path, fname)

        xdata, ydata = self.plot_data
        if self.plotMethod in ['Average', 'Sum']:
            if self.plotMethod == 'Average':
                s_ydata = np.nanmean(ydata, 0)
            else:
                s_ydata = np.nansum(ydata, 0)

            # Write to xye
            xye_fname = f'{fname}.xye'
            ut.write_xye(xye_fname, xdata, s_ydata)

            # Write to csv
            csv_fname = f'{fname}.csv'
            ut.write_csv(csv_fname, xdata, s_ydata)

        idxs = [arch.replace(f'{self.sphere.name}_', '') for arch in self.arch_names]
        for nn, (s_ydata, idx) in enumerate(zip(ydata, idxs)):
            # Write to xye
            xye_fname = f'{fname}_{str(idx).zfill(4)}.xye'
            ut.write_xye(xye_fname, xdata, s_ydata)

            # Write to csv
            csv_fname = f'{fname}_{str(idx).zfill(4)}.csv'
            ut.write_csv(csv_fname, xdata, s_ydata)

        if not auto:
            scene = self.plot_viewBox.scene()
            exporter = pyqtgraph.exporters.ImageExporter(scene)
            h = exporter.params.param('height').value()
            w = exporter.params.param('width').value()
            h_new = 600
            w_new = int(np.round(w/h * h_new, 0))
            exporter.params.param('height').setValue(h_new)
            exporter.params.param('width').setValue(w_new)
            exporter.export(fname + '.png')

    def get_colors(self):
        # Define color tuples

        colors = (1, 1, 1)
        if self.cmap == 'Default':
            colors_tuples = [cm.get_cmap('tab10'), cm.get_cmap('Set3'), cm.get_cmap('tab20b', 5)]
            for nn, color_tuples in enumerate(colors_tuples):
                if nn == 0:
                    colors = np.asarray(color_tuples.colors)
                else:
                    colors = np.vstack((colors, np.asarray(color_tuples.colors)[:, 0:3]))

            colors_tuples = cm.get_cmap('jet')
            more_colors = colors_tuples(np.linspace(0, 1, len(self.arch_names)))
            colors = np.vstack((colors, more_colors[:, 0:3]))

        else:
            try:
                colors_tuples = cm.get_cmap(self.cmap)
            except ValueError:
                colors_tuples = cm.get_cmap('jet', 256)
            colors = colors_tuples(np.linspace(0, 1, len(self.arch_names)))[:, 0:3]

        # colors = np.round(colors * [255, 255, 255, 1]).astype(int)
        colors = np.round(colors * [255, 255, 255]).astype(int)
        colors = [tuple(color[:3]) for color in colors]

        return colors

    def get_profile_chi(self, arch):
        """
        Args:
            arch (EwaldArch Object):

        Returns:
            intensity (ndarray): Intensity integrated along Chi
                                 over a range of Q specified by UI
        """
        pass

    def get_chi_1d(self, arch):
        """
        Args:
            arch (EwaldArch Object):

        Returns:
            intensity (ndarray): Intensity integrated along Chi
                                 over a range of Q specified by UI
        """
        pass

    def trackMouse(self):
        """
        Sets up mouse trancking
        Returns: x, y (z) coordinates on screen

        """
        proxy = pg.SignalProxy(signal=self.plot.scene().sigMouseMoved, rateLimit=60, slot=self.mouseMoved)
        proxy.signal.connect(self.mouseMoved)

    def mouseMoved(self, pos):
        if len(self.curves) == 0:
            return

        if self.plot.sceneBoundingRect().contains(pos):
            vb = self.plot.vb
            self.plot_win.setCursor(pyQt.CrossCursor)
            mousePoint = vb.mapSceneToView(pos)
            self.pos_label.setText(f'x={mousePoint.x():.2f}, y={mousePoint.y():.2e}')
        else:
            self.pos_label.setText('')
            self.plot_win.setCursor(pyQt.ArrowCursor)

    def update_wf_pmesh(self):
        """Updates data in 1D plot Frame
        """
        self.setup_wf_layout()

        xdata_, data_ = self.plot_data
        s_xdata, data = xdata_.copy(), data_.copy()
        data = data[self.wf_start::self.wf_step, :]

        x_max, x_min = np.max(s_xdata), np.min(s_xdata)
        x_step = (x_max - x_min)/len(s_xdata)
        s_xdata = np.append(s_xdata, [x_max + x_step])
        s_xdata -= x_step/2
        s_xdata = np.tile(s_xdata, (data.shape[0]+1, 1)).T

        # Set YAxis Unit
        if self.wf_yaxis == 'Frame #':
            s_ydata = np.asarray(np.arange(data.shape[0]) + self.wf_start + 1, dtype=float)
        else:
            # TODO: Fix below for more WF options
            if self.wf_yaxis == 'Time (s)':
                s_ydata = np.asarray([self.data_1d[idx].scan_info['epoch'] for idx in self.idxs])
                s_ydata -= s_ydata.min()
            elif self.wf_yaxis == 'Time (minutes)':
                s_ydata = np.asarray([self.data_1d[idx].scan_info['epoch'] for idx in self.idxs])/60.
                s_ydata -= s_ydata.min()
            else:
                s_ydata = np.asarray([self.data_1d[idx].scan_info[self.wf_yaxis] for idx in self.idxs])

            s_ydata = s_ydata[self.wf_start::self.wf_step]

        y_max, y_min = np.max(s_ydata), np.min(s_ydata)
        y_step = (y_max - y_min)/len(s_ydata)
        s_ydata = np.append(s_ydata, [y_max + y_step])
        s_ydata -= y_step/2.
        s_ydata = np.tile(s_ydata, (data.shape[1]+1, 1))

        # self.wf_widget.setImage(data.T, scale=self.scale, cmap=self.cmap)
        levels = np.nanpercentile(data, (1, 98))
        self.wf_widget.imageItem.setLevels(levels)
        self.wf_widget.imageItem.setData(s_xdata, s_ydata, data.T)
        self.wf_widget.imageItem.informViewBoundsChanged()

        # rect = get_rect(s_xdata, np.arange(data.shape[0]))
        rect = get_rect(s_xdata[:, 0], s_ydata[0])
        self.wf_widget.setRect(rect)

        plotUnit = self.ui.plotUnit.currentIndex()
        self.wf_widget.image_plot.setLabel("bottom", x_labels_1D[plotUnit],
                                           units=x_units_1D[plotUnit])
        self.wf_widget.image_plot.setLabel("left", self.wf_yaxis)

        return data



from collections import namedtuple
from dataclasses import dataclass, field
import copy
import tempfile

import numpy as np
from pyFAI import units
import h5py

from .nzarrays import nzarray1d, nzarray2d
from .. import _utils as utils


def parse_unit(result, wavelength):
    """Helper function to take integrator result and return a two
    theta and q array regardless of the unit used for integration.

    args:
        result: result from 1dintegrator
        wavelength: wavelength for conversion in meters

    returns:
        int_1d_2theta: two theta array
        int_1d_q: q array
    """
    if wavelength is None:
        return result.radial, None

    if result.unit == units.TTH_DEG or str(result.unit) == '2th_deg':
        int_1d_2theta = result.radial
        int_1d_q = (
            (4 * np.pi / (wavelength*1e10)) *
            np.sin(np.radians(int_1d_2theta / 2))
        )
    elif result.unit == units.Q_A or str(result.unit) == 'q_A^-1':
        int_1d_q = result.radial
        int_1d_2theta = (
            2*np.degrees(
                np.arcsin(
                    int_1d_q *
                    (wavelength * 1e10) /
                    (4 * np.pi)
                )
            )
        )
    # TODO: implement other unit options for unit
    return int_1d_2theta, int_1d_q


class int_1d_data:
    """Container for 1-dimensional integration data returned by pyFAI,
    scanning detector version.
    
    attributes:
        raw: nzarray1d, raw integrated signal
        pcount: nzarray1d, how many pixels in each bin
        norm: nzarray1d, integrated signal normalized by number of
            pixels
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
    
    methods:
        from_hdf5: Loads in data from an hdf5 file
        from_result: Parses data in from result returned by pyFAI
        parse_unit: Ensures both q and ttheta data is held
        to_hdf5: Saves data to hdf5 file
    """
    def __init__(self, raw=None, pcount=None, norm=None, ttheta=0, q=0,
                 sigma=None, sigma_raw=None, monitor=1):
        """
        raw: nzarray1d, raw integrated signal
        pcount: nzarray1d, how many pixels in each bin
        norm: nzarray1d, integrated signal normalized by number of
            pixels
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
        """
        self.raw = raw
        self.pcount = pcount
        self.norm = norm
        self.ttheta = ttheta
        self.q = q
        self.sigma = sigma
        self.sigma_raw = sigma_raw

    def from_result(self, result, wavelength, monitor=1, unit=None):
        """Parses out result obtained by pyFAI AzimuthalIntegrator.
        
        args:
            result: object returned by AzimuthalIntegrator
            wavelength: float, energy of the beam in meters
        """
        self.ttheta, self.q = self.parse_unit(
            result, wavelength, unit)

        self.pcount = result._count
        self.raw = result._sum_signal/monitor
        self.norm = self.raw/self.pcount
        if result.sigma is None:
            self.sigma = result._sum_signal
            self.sigma.data = np.sqrt(self.sigma.data)
            self.sigma /= (self.pcount * monitor)
            self.sigma_raw = result._sum_signal / (monitor**2)
        else:
            self.sigma = result.sigma / monitor
            self.sigma_raw = ((result._count * result.sigma)**2) / (monitor**2)
    
    def parse_unit(self, result, wavelength, unit=None):
        """Helper function to take integrator result and return a two
        theta and q array regardless of the unit used for integration.

        args:
            result: result from 1dintegrator
            wavelength: wavelength for conversion in meters

        returns:
            int_1d_2theta: two theta array
            int_1d_q: q array
        """
        if wavelength is None:
            return result.radial, None

        if unit is None:
            unit = result.unit

        if unit == units.TTH_DEG or str(unit) == '2th_deg':
            int_1d_2theta = result.radial
            int_1d_q = (
                (4 * np.pi / (wavelength*1e10)) *
                np.sin(np.radians(int_1d_2theta / 2))
            )
        elif unit == units.Q_A or str(unit) == 'q_A^-1':
            int_1d_q = result.radial
            int_1d_2theta = (
                2*np.degrees(
                    np.arcsin(
                        int_1d_q *
                        (wavelength * 1e10) /
                        (4 * np.pi)
                    )
                )
            )
        # TODO: implement other unit options for unit
        return int_1d_2theta, int_1d_q

    def to_hdf5(self, grp, compression=None):
        """Saves data to hdf5 file.
        
        args:
            grp: h5py Group or File, where the data will be saved
            compression: str, compression algorithm to use. See h5py
                documentation.
        """
        for key in ['raw', 'pcount', 'norm', 'sigma', 'sigma_raw']:
            attr = self.__getattribute__(key)
            if key in grp:
                if not utils.check_encoded(grp, 'nzarray'):
                    del grp[key]
                    sgrp = grp.create_group(key)
                else:
                    sgrp = grp[key]
            else:
                sgrp = grp.create_group(key)
            attr.to_hdf5(sgrp, compression)

        utils.attributes_to_h5(self, grp, ['ttheta', 'q'], compression=compression)
    
    def from_hdf5(self, grp):
        """Loads in data from hdf5 file.
        
        args:
            grp: h5py Group or File, object to load data from.
        """
        self.raw.from_hdf5(grp['raw'])
        self.pcount.from_hdf5(grp['pcount'])
        self.norm.from_hdf5(grp['norm'])
        try:
            self.sigma.from_hdf5(grp['sigma'])
            self.sigma_raw.from_hdf5(grp['sigma_raw'])
        except KeyError:
            self.sigma = copy.deepcopy(self.norm)
            data = self.sigma.data
            if data[data > 0].size > 0:
                minval = data[data > 0].min()
                data[data > 0] = np.sqrt(data[data > 0]/minval) * minval
            self.sigma_raw = self.sigma * self.pcount
        utils.h5_to_attributes(self, grp, ['ttheta', 'q'])
    
    def __setattr__(self, name, value):
        """Ensures raw, norm, and pcount are nzarray1d objects.
        """
        if name in ['raw', 'norm', 'pcount', 'sigma', 'sigma_raw']:
            self.__dict__[name] = nzarray1d(value)
        else:
            super().__setattr__(name, value)
    
    def __add__(self, other):
        out = self.__class__()
        out.raw = self.raw + other.raw
        out.pcount = self.pcount + other.pcount
        out.sigma_raw = self.sigma_raw + other.sigma_raw

        #out.sigma = self.sigma*self.sigma + other.sigma*other.sigma
        #out.sigma.data = np.sqrt(out.sigma.data)
        out.sigma = copy.deepcopy(out.sigma_raw)
        out.sigma.data = np.sqrt(out.sigma_raw.data)
        out.sigma = out.sigma/out.pcount

        out.norm = out.raw/out.pcount
        out.ttheta = copy.deepcopy(self.ttheta)
        out.q = copy.deepcopy(self.q)
        try:
            out.chi = copy.deepcopy(self.chi)
        except AttributeError:
            pass
        return out

class int_2d_data(int_1d_data):
    """Container for 2-dimensional integration data returned by pyFAI.
    
    attributes:
        raw: nzarray2d, raw integrated signal
        pcount: nzarray2d, how many pixels in each bin
        norm: nzarray2d, integrated signal normalized by number of
            pixels
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
        chi: numpy array, chi values
    
    methods:
        from_hdf5: Loads in data from an hdf5 file
        from_result: Parses data in from result returned by pyFAI
        parse_unit: Ensures both q and ttheta data is held
        to_hdf5: Saves data to hdf5 file
    """
    def __init__(self,  raw=None, pcount=None, norm=None, ttheta=0, q=0,
                 chi=0, sigma=None, sigma_raw=None):
        """
        raw: nzarray2d, raw integrated signal
        pcount: nzarray2d, how many pixels in each bin
        norm: nzarray2d, integrated signal normalized by number of
            pixels
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
        chi: numpy array, chi values
        """
        self.raw = raw
        self.pcount = pcount
        self.norm = norm
        self.ttheta = ttheta
        self.q = q
        self.chi = chi
        self.sigma = sigma
        self.sigma_raw = sigma_raw

    def from_result(self, result, wavelength, monitor=1, unit=None):
        """Parses out result obtained by pyFAI AzimuthalIntegrator.
        
        args:
            result: object returned by AzimuthalIntegrator
            wavelength: float, energy of the beam in meters
        """
        super(int_2d_data, self).from_result(result, wavelength, monitor, unit)
        self.chi = result.azimuthal
    
    def from_hdf5(self, grp):
        """Loads in data from hdf5 file.
        
        args:
            grp: h5py Group or File, object to load data from.
        """
        super().from_hdf5(grp)
        utils.h5_to_attributes(self, grp, ['chi'])

    def to_hdf5(self, grp, compression=None):
        """Saves data to hdf5 file.
        
        args:
            grp: h5py Group or File, where the data will be saved
            compression: str, compression algorithm to use. See h5py
                documentation.
        """
        super().to_hdf5(grp, compression)
        utils.attributes_to_h5(self, grp, ['chi'], compression=compression)

    def __setattr__(self, name, value):
        """Ensures raw, norm, and pcount are nzarray2d objects.
        """
        if name in ['raw', 'norm', 'pcount', 'sigma', 'sigma_raw']:
            self.__dict__[name] = nzarray2d(value)
        else:
            super().__setattr__(name, value)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'maskWidgetUI.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(363, 366)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.toolsFrame = QtWidgets.QFrame(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolsFrame.sizePolicy().hasHeightForWidth())
        self.toolsFrame.setSizePolicy(sizePolicy)
        self.toolsFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.toolsFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.toolsFrame.setObjectName("toolsFrame")
        self.gridLayout = QtWidgets.QGridLayout(self.toolsFrame)
        self.gridLayout.setObjectName("gridLayout")
        self.lineMaskButton = QtWidgets.QPushButton(self.toolsFrame)
        self.lineMaskButton.setObjectName("lineMaskButton")
        self.gridLayout.addWidget(self.lineMaskButton, 3, 0, 1, 1)
        self.frame = QtWidgets.QFrame(self.toolsFrame)
        self.frame.setMinimumSize(QtCore.QSize(150, 150))
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.frame)
        self.gridLayout_2.setContentsMargins(1, 1, 1, 1)
        self.gridLayout_2.setSpacing(1)
        self.gridLayout_2.setObjectName("gridLayout_2")
        spacerItem = QtWidgets.QSpacerItem(88, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem, 2, 2, 1, 1)
        self.xMinBox = QtWidgets.QSpinBox(self.frame)
        self.xMinBox.setObjectName("xMinBox")
        self.gridLayout_2.addWidget(self.xMinBox, 3, 2, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(88, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem1, 3, 1, 1, 1)
        self.yMinBox = QtWidgets.QSpinBox(self.frame)
        self.yMinBox.setObjectName("yMinBox")
        self.gridLayout_2.addWidget(self.yMinBox, 2, 1, 1, 1)
        self.xMaxBox = QtWidgets.QSpinBox(self.frame)
        self.xMaxBox.setObjectName("xMaxBox")
        self.gridLayout_2.addWidget(self.xMaxBox, 1, 2, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(88, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem2, 3, 3, 1, 1)
        spacerItem3 = QtWidgets.QSpacerItem(88, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem3, 1, 1, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(88, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem4, 1, 3, 1, 1)
        self.yMaxBox = QtWidgets.QSpinBox(self.frame)
        self.yMaxBox.setObjectName("yMaxBox")
        self.gridLayout_2.addWidget(self.yMaxBox, 2, 3, 1, 1)
        self.label = QtWidgets.QLabel(self.frame)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 4, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.frame)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 0, 4, 1, 1)
        self.yMaxSlider = QtWidgets.QSlider(self.frame)
        self.yMaxSlider.setOrientation(QtCore.Qt.Vertical)
        self.yMaxSlider.setObjectName("yMaxSlider")
        self.gridLayout_2.addWidget(self.yMaxSlider, 1, 4, 3, 1)
        self.yMinSlider = QtWidgets.QSlider(self.frame)
        self.yMinSlider.setOrientation(QtCore.Qt.Vertical)
        self.yMinSlider.setObjectName("yMinSlider")
        self.gridLayout_2.addWidget(self.yMinSlider, 1, 0, 3, 1)
        self.xMaxSlider = QtWidgets.QSlider(self.frame)
        self.xMaxSlider.setOrientation(QtCore.Qt.Horizontal)
        self.xMaxSlider.setObjectName("xMaxSlider")
        self.gridLayout_2.addWidget(self.xMaxSlider, 0, 1, 1, 3)
        self.xMinSlider = QtWidgets.QSlider(self.frame)
        self.xMinSlider.setOrientation(QtCore.Qt.Horizontal)
        self.xMinSlider.setObjectName("xMinSlider")
        self.gridLayout_2.addWidget(self.xMinSlider, 4, 1, 1, 3)
        self.gridLayout.addWidget(self.frame, 4, 0, 1, 2)
        self.ellipseMaskButton = QtWidgets.QPushButton(self.toolsFrame)
        self.ellipseMaskButton.setObjectName("ellipseMaskButton")
        self.gridLayout.addWidget(self.ellipseMaskButton, 2, 1, 1, 1)
        self.addROI = QtWidgets.QRadioButton(self.toolsFrame)
        self.addROI.setChecked(True)
        self.addROI.setObjectName("addROI")
        self.gridLayout.addWidget(self.addROI, 1, 0, 1, 1)
        self.rectMaskButton = QtWidgets.QPushButton(self.toolsFrame)
        self.rectMaskButton.setObjectName("rectMaskButton")
        self.gridLayout.addWidget(self.rectMaskButton, 2, 0, 1, 1)
        self.subtractROI = QtWidgets.QRadioButton(self.toolsFrame)
        self.subtractROI.setObjectName("subtractROI")
        self.gridLayout.addWidget(self.subtractROI, 1, 1, 1, 1)
        self.polyMaskButton = QtWidgets.QPushButton(self.toolsFrame)
        self.polyMaskButton.setObjectName("polyMaskButton")
        self.gridLayout.addWidget(self.polyMaskButton, 3, 1, 1, 1)
        self.archList = QtWidgets.QComboBox(self.toolsFrame)
        self.archList.setObjectName("archList")
        self.gridLayout.addWidget(self.archList, 0, 0, 1, 1)
        self.clearButton = QtWidgets.QPushButton(self.toolsFrame)
        self.clearButton.setObjectName("clearButton")
        self.gridLayout.addWidget(self.clearButton, 0, 1, 1, 1)
        self.setCurrent = QtWidgets.QPushButton(self.toolsFrame)
        self.setCurrent.setObjectName("setCurrent")
        self.gridLayout.addWidget(self.setCurrent, 5, 0, 1, 1)
        self.setGlobal = QtWidgets.QPushButton(self.toolsFrame)
        self.setGlobal.setObjectName("setGlobal")
        self.gridLayout.addWidget(self.setGlobal, 5, 1, 1, 1)
        self.horizontalLayout.addWidget(self.toolsFrame)

        self.retranslateUi(Form)
        self.xMaxBox.valueChanged['int'].connect(self.xMaxSlider.setValue)
        self.xMaxSlider.sliderMoved['int'].connect(self.xMaxBox.setValue)
        self.yMaxBox.valueChanged['int'].connect(self.yMaxSlider.setValue)
        self.yMaxSlider.sliderMoved['int'].connect(self.yMaxBox.setValue)
        self.xMinBox.valueChanged['int'].connect(self.xMinSlider.setValue)
        self.xMinSlider.sliderMoved['int'].connect(self.xMinBox.setValue)
        self.yMinSlider.sliderMoved['int'].connect(self.yMinBox.setValue)
        self.yMinBox.valueChanged['int'].connect(self.yMinSlider.setValue)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.lineMaskButton.setText(_translate("Form", "Line"))
        self.label.setText(_translate("Form", "Min"))
        self.label_2.setText(_translate("Form", "Max"))
        self.ellipseMaskButton.setText(_translate("Form", "Ellipse"))
        self.addROI.setText(_translate("Form", "Add"))
        self.rectMaskButton.setText(_translate("Form", "Rect"))
        self.subtractROI.setText(_translate("Form", "Subtract"))
        self.polyMaskButton.setText(_translate("Form", "Poly"))
        self.clearButton.setText(_translate("Form", "Clear"))
        self.setCurrent.setText(_translate("Form", "Set Current"))
        self.setGlobal.setText(_translate("Form", "Set Global"))



# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports
from queue import Queue
import multiprocessing as mp
import copy
import os

# Other imports

# Qt imports
from pyqtgraph.Qt import QtWidgets, QtCore

# This module imports
from xdart.modules.ewald import EwaldSphere, EwaldArch
from xdart.utils import catch_h5py_file as catch
from xdart import utils as ut
from .tthetaUI import Ui_Form
from .h5viewer import H5Viewer
from .display_frame_widget import displayFrameWidget
from .integrator import integratorTree
from .sphere_threads import integratorThread
from .metadata import metadataWidget
from .wranglers import specWrangler, liveSpecWrangler, wranglerWidget

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

QWidget = QtWidgets.QWidget
QSizePolicy = QtWidgets.QSizePolicy
QFileDialog = QtWidgets.QFileDialog

wranglers = {
    'SPEC': specWrangler, 
    'Live Spec': liveSpecWrangler
}


def spherelocked(func):
    def wrapper(self, *args, **kwargs):
        if isinstance(self.sphere, EwaldSphere):
            with self.sphere.sphere_lock:
                func(self, *args, **kwargs)
                return func(self, *args, **kwargs)
    return wrapper


class tthetaWidget(QWidget):
    """Tab for integrating data collected by a scanning area detector.
    As of current version, only handles a single angle (2-theta).
    Displays raw images, stitched Q Chi arrays, and integrated I Q
    arrays. Also displays metadata and exposes parameters for
    controlling integration.
    
    children:
        displayframe: widget which handles displaying images and
            plotting data.
        h5viewer: Has a file explorer panel for loading scans, and
            a panel which shows images that are associated with the
            loaded scan. Has other file saving and loading functions
            as well as configuration saving and loading functions.
        integrator_thread: Not visible to user, but a sub-thread which
            handles integration to free resources for the gui
        integratorTree: Widget for setting the basic integration
            parameters. Also has buttons for starting integration.
        metawidget: Table wiget which displays metadata either for
            entire scan or individual image.
        
    attributes:
        arch: EwaldArch, currently loaded arch object
        command_queue: Queue, used to send commands to wrangler
        dirname: str, absolute path of current directory for scan
        file_lock: mp.Condition, process safe lock
        fname: str, current data file name
        sphere: EwaldSphere, current scan data
        timer: QTimer, currently unused but can be used for periodic
            functions.
        ui: Ui_Form, layout from qtdesigner
    
    methods:
        bai_1d: Sends signal to thread to start integrating 1d
        bai_2d:  Sends signal to thread to start integrating 2d
        clock: Unimplemented, used for periodic updates
        close: Handles cleanup prior to closing
        enable_integration: Sets enabled status of widgets related to
            integration
        first_arch, last_arch, next_arch: Handle moving between
            different arches in the overall sphere
        load_and_set: Combination of load and set methods. Also governs
            file explorer behavior in h5viewer.
        load_sphere: 
    """
    def __init__(self, local_path=None, parent=None):
        # ic()
        super().__init__(parent)

        # Data object initialization
        self.file_lock = mp.Condition()
        self.local_path = local_path
        dirname = os.path.join(local_path)
        if not os.path.isdir(dirname):
            os.mkdir(dirname)
        self.fname = os.path.join(dirname, 'default.hdf5')
        self.sphere = EwaldSphere('null_main', data_file=self.fname)
        self.sphere.file_lock = self.file_lock
        self.arch = EwaldArch()

        self.ui = Ui_Form()
        self.ui.setupUi(self)

        # H5Viewer setup
        self.h5viewer = H5Viewer(self.file_lock, self.local_path, dirname,
                                 self.sphere, self.arch, self.ui.hdf5Frame)
        self.h5viewer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.ui.hdf5Frame.setLayout(self.h5viewer.layout)
        self.h5viewer.ui.listData.addItem('No data')
        self.h5viewer.ui.listData.setCurrentRow(0)

        # H5Viewer signal connections
        self.h5viewer.sigUpdate.connect(self.set_data)
        self.h5viewer.file_thread.sigTaskStarted.connect(self.thread_state_changed)
        self.h5viewer.sigThreadFinished.connect(self.thread_state_changed)
        self.h5viewer.ui.listData.itemClicked.connect(self.enable_last)

        # DisplayFrame setup
        self.displayframe = displayFrameWidget(self.sphere, self.arch, 
                                               parent=self.ui.middleFrame)
        self.ui.middleFrame.setLayout(self.displayframe.ui.layout)

        # DisplayFrame signal connections
        self.displayframe.ui.pushRight.clicked.connect(self.next_arch)
        self.displayframe.ui.pushLeft.clicked.connect(self.prev_arch)
        self.displayframe.ui.pushRightLast.clicked.connect(self.last_arch)
        self.displayframe.ui.pushLeftLast.clicked.connect(self.first_arch)
        self.h5viewer.actionSaveImage.triggered.connect(
            self.displayframe.save_image
        )
        self.h5viewer.actionSaveArray.triggered.connect(
            self.displayframe.save_array
        )
        
        # IntegratorFrame setup
        self.integratorTree = integratorTree(self.sphere, self.arch,
                                             self.file_lock)
        self.ui.integratorFrame.setLayout(self.integratorTree.ui.verticalLayout)
        self.integratorTree.update()

        # Integrator signal connections
        self.integratorTree.integrator_thread.started.connect(self.thread_state_changed)
        self.integratorTree.integrator_thread.update.connect(self.integrator_thread_update)
        self.integratorTree.integrator_thread.finished.connect(self.integrator_thread_finished)

        # Metadata setup
        self.metawidget = metadataWidget(self.sphere, self.arch)
        self.ui.metaFrame.setLayout(self.metawidget.layout)

        # Wrangler frame setup
        self.wrangler = wranglerWidget("uninitialized", mp.Condition())
        for name, w in wranglers.items():
            self.ui.wranglerStack.addWidget(
                w(
                    self.fname,
                    self.file_lock
                )
            )
            self.ui.wranglerBox.addItem(name)
        self.ui.wranglerStack.currentChanged.connect(self.set_wrangler)
        self.command_queue = Queue()
        self.set_wrangler(self.ui.wranglerStack.currentIndex())
        
        self.integratorTree.get_args('bai_1d')
        self.integratorTree.get_args('bai_2d')
        
        # Setup defaultWidget in h5viewer with parameters
        parameters = [self.integratorTree.parameters]
        for i in range(self.ui.wranglerStack.count()):
            w = self.ui.wranglerStack.widget(i)
            parameters.append(w.parameters)
        self.h5viewer.defaultWidget.set_parameters(parameters)
        self.h5viewer.load_starting_defaults()
        self.show()

        self.resize(1300, self.height())
        self.ui.mainSplitter.setSizes([200,700,400])
        """
        self.timer = Qt.QtCore.QTimer()
        self.timer.timeout.connect(self.clock)
        self.timer.start()
    """
    def set_wrangler(self, qint):
        """Sets the wrangler based on the selected item in the dropdown.
        Syncs the wrangler's attributes and wires signals as needed.
        
        args:
            qint: Qt int, index of the new wrangler
        """
        # ic()
        if 'wrangler' in self.__dict__:
            self.disconnect_wrangler()
            
        self.wrangler = self.ui.wranglerStack.widget(qint)
        self.wrangler.input_q = self.command_queue
        self.wrangler.fname = self.fname
        self.wrangler.file_lock = self.file_lock
        self.wrangler.sigStart.connect(self.start_wrangler)
        self.wrangler.sigUpdateData.connect(self.update_data)
        self.wrangler.sigUpdateFile.connect(self.new_scan)
        self.wrangler.started.connect(self.thread_state_changed)
        self.wrangler.finished.connect(self.wrangler_finished)
        self.wrangler.setup()
        self.h5viewer.sigNewFile.connect(self.wrangler.set_fname)
    
    def disconnect_wrangler(self):
        """Disconnects all signals attached the the current wrangler
        """
        # ic()
        for signal in (self.wrangler.sigStart,
                       self.wrangler.sigUpdateData,
                       self.wrangler.sigUpdateFile,
                       self.wrangler.finished,
                       self.h5viewer.sigNewFile):
            while True:
                try:
                    signal.disconnect()
                except TypeError:
                    break

    def thread_state_changed(self):
        """Called whenever a thread is started or finished.
        """
        # ic()
        wrangler_running = self.wrangler.thread.isRunning()
        integrator_running = self.integratorTree.integrator_thread.isRunning()
        loader_running = self.h5viewer.file_thread.running
        same_name = self.sphere.name == self.wrangler.scan_name

        if loader_running:
            self.h5viewer.ui.listData.setEnabled(False)
            self.h5viewer.ui.listScans.setEnabled(False)
            self.h5viewer.set_open_enabled(False)
            self.integratorTree.setEnabled(False)
            self.displayframe.ui.setMaskButton.setEnabled(False)

        elif integrator_running:
            self.h5viewer.ui.listData.setEnabled(True)
            self.integratorTree.setEnabled(False)
            self.displayframe.ui.setMaskButton.setEnabled(False)
            self.h5viewer.ui.listScans.setEnabled(False)
            self.h5viewer.set_open_enabled(False)
            if same_name or wrangler_running:
                self.wrangler.enabled(False)
            else:
                self.wrangler.enabled(True)

        elif wrangler_running:
            self.h5viewer.ui.listData.setEnabled(True)
            self.h5viewer.ui.listScans.setEnabled(True)
            self.h5viewer.set_open_enabled(True)
            self.h5viewer.paramMenu.setEnabled(False)
            self.wrangler.enabled(False)
            if same_name:
                self.integratorTree.setEnabled(False)
                self.displayframe.ui.setMaskButton.setEnabled(False)
            else:
                self.integratorTree.setEnabled(True)

        else:
            self.h5viewer.ui.listData.setEnabled(True)
            self.h5viewer.ui.listScans.setEnabled(True)
            self.h5viewer.set_open_enabled(True)
            self.integratorTree.setEnabled(True)
            self.displayframe.ui.setMaskButton.setEnabled(True)
            self.wrangler.enabled(True)

    def update_data(self, q):
        """Called by signal from wrangler. If the current scan name
        is the same as the wrangler scan name, updates the data in
        memory.
        """
        # ic()
        with self.sphere.sphere_lock:
            if self.sphere.name == self.wrangler.scan_name:
                self.h5viewer.file_thread.queue.put("update_sphere")

                with self.file_lock:
                    self.update_all()

    def enable_last(self, q):
        """
        Parameters
        ----------
        q : Qt.QtWidgets.QListWidgetItem
        """
        # ic()
        self.displayframe.auto_last = False
        self.displayframe.ui.pushRightLast.setEnabled(True)

    def set_data(self):
        """Connected to h5viewer, sets the data in displayframe based
        on the selected image or overall data.
        """
        # ic()
        if self.sphere.name != 'null_main':
            self.displayframe.update()
            
            if self.arch.idx is None:
                self.displayframe.ui.imageIntRaw.setEnabled(False)
                self.displayframe.ui.imageMask.setEnabled(False)

                self.integratorTree.ui.all1D.setChecked(True)
                self.integratorTree.ui.all1D.setEnabled(False)
                self.integratorTree.ui.all2D.setChecked(True)
                self.integratorTree.ui.all2D.setEnabled(False)

            else:
                self.displayframe.ui.imageIntRaw.setEnabled(True)
                self.displayframe.ui.imageMask.setEnabled(True)

                self.integratorTree.ui.all1D.setEnabled(True)
                self.integratorTree.ui.all2D.setEnabled(True)

            self.metawidget.update()
            self.integratorTree.update()
    
    def next_arch(self):
        """Advances to next arch in data list, updates displayframe
        """
        # ic()
        if (self.arch == self.sphere.arches.iloc(-1).idx or
            self.arch is None or
            self.h5viewer.ui.listData.currentRow() == \
                self.h5viewer.ui.listData.count() - 1):
            pass
        else:
            self.h5viewer.ui.listData.setCurrentRow(
                self.h5viewer.ui.listData.currentRow() + 1
            )
            self.displayframe.auto_last = False
            self.displayframe.ui.pushRightLast.setEnabled(True)
    
    def prev_arch(self):
        """Goes back one arch in data list, updates displayframe
        """
        # ic()
        if (self.arch == self.sphere.arches.iloc(0).idx or
            self.arch.idx is None or
            self.h5viewer.ui.listData.currentRow() == 1):
            pass
        else:
            self.h5viewer.ui.listData.setCurrentRow(
                self.h5viewer.ui.listData.currentRow() - 1
            )
            self.displayframe.auto_last = False
            self.displayframe.ui.pushRightLast.setEnabled(True)
    
    def last_arch(self):
        """Advances to last arch in data list, updates displayframe, and
        set auto_last to True
        """
        # ic()
        if self.arch.idx is None:
            pass

        else: 
            if self.arch.idx == self.sphere.arches.index[-1]:
                pass

            else:
                self.h5viewer.ui.listData.setCurrentRow(
                    self.h5viewer.ui.listData.count() - 1
                )
        
            self.displayframe.auto_last = True
            self.displayframe.ui.pushRightLast.setEnabled(False)

    def first_arch(self):
        """Goes to first arch in data list, updates displayframe
        """
        # ic()
        if self.arch == self.sphere.arches.iloc(0).idx or self.arch.idx is None:
            pass
        else:
            self.h5viewer.ui.listData.setCurrentRow(1)
            self.displayframe.auto_last = False
            self.displayframe.ui.pushRightLast.setEnabled(True)
    
    def close(self):
        """Tries a graceful close.
        """
        # ic()
        del(self.sphere)
        del(self.displayframe.sphere)
        del(self.arch)
        del(self.displayframe.arch)
        super().close()
    
    def enable_integration(self, enable=True):
        """Calls the integratorTree setEnabled function.
        """
        # ic()
        self.integratorTree.setEnabled(enable)

    def update_all(self):
        """Updates all data in displays
        TODO: Currently taking the most time for the main gui thread
        """
        # ic()
        self.h5viewer.update_data()
        if self.displayframe.auto_last:
            self.h5viewer.ui.listData.setCurrentRow(
                self.h5viewer.ui.listData.count() - 1
            )
        else:
            self.displayframe.update()
        self.metawidget.update()
    
    def integrator_thread_update(self, idx):
        self.thread_state_changed()
        if self.displayframe.auto_last:
            items = self.h5viewer.ui.listData.findItems(str(idx),
                                                        QtCore.Qt.MatchExactly)
            for item in items:
                self.h5viewer.ui.listData.setCurrentItem(item)
            self.displayframe.auto_last = True
        else:
            self.displayframe.update()
    
    def integrator_thread_finished(self):
        """Function connected to threadFinished signals for
        integratorThread
        """
        # ic()
        self.thread_state_changed()
        if self.integratorTree.integrator_thread.method in ["bai_1d_SI", "bai_2d_SI"]:
            self.displayframe.update()
        else:
            self.update_all()
        if not self.wrangler.thread.isRunning():
            self.ui.wranglerBox.setEnabled(True)
            self.wrangler.enabled(True)
    
    def new_scan(self, name, fname):
        """Connected to sigUpdateFile from wrangler. Called when a new
        scan is started. 
        
        args:
            name: str, scan name
            fname: str, path to data file for scan
        """
        # ic()
        if self.sphere.name == name or self.sphere.name == 'null_main':
            self.h5viewer.dirname = os.path.dirname(fname)
            self.h5viewer.set_file(fname)
        self.h5viewer.update()

    def start_wrangler(self):
        """Sets up wrangler, ensures properly synced args, and starts
        the wrangler.thread main method.
        """
        # ic()
        self.ui.wranglerBox.setEnabled(False)
        self.wrangler.enabled(False)
        args = {'bai_1d_args': self.sphere.bai_1d_args,
                'bai_2d_args': self.sphere.bai_2d_args}
        self.wrangler.sphere_args = copy.deepcopy(args)
        self.wrangler.setup()
        self.wrangler.thread.start()
    
    def wrangler_finished(self):
        """Called by the wrangler finished signal. If current scan
        matches the wrangler scan, allows for integration.
        """
        # ic()
        self.thread_state_changed()
        if self.sphere.name == self.wrangler.scan_name:
            self.integrator_thread_finished()
        else:
            self.ui.wranglerBox.setEnabled(True)
            self.wrangler.enabled(True)
